from .base import BaseExtensionManager
from .nbextensions import NbExtensionManager
from .serverextensions import ServerExtensionManager


class E2xExtensionManager(BaseExtensionManager):
    def __init__(self):
        super().__init__()
        self.managers = [NbExtensionManager(), ServerExtensionManager()]

    def deactivate(self, sys_prefix: bool = True, user: bool = False) -> None:
        for manager in self.managers:
            manager.deactivate(sys_prefix=sys_prefix, user=user)

    def activate_teacher(self, sys_prefix: bool = True, user: bool = False) -> None:
        for manager in self.managers:
            manager.activate_teacher(sys_prefix=sys_prefix, user=user)

    def activate_student(self, sys_prefix: bool = True, user: bool = False) -> None:
        for manager in self.managers:
            manager.activate_student(sys_prefix=sys_prefix, user=user)

    def activate_student_exam(
        self, sys_prefix: bool = True, user: bool = False
    ) -> None:
        for manager in self.managers:
            manager.activate_student_exam(sys_prefix=sys_prefix, user=user)



import unittest

from e2xauthoring.managers import PresetManager
from nbformat.v4 import new_markdown_cell, new_notebook
from nbgrader.nbgraderformat import ValidationError

from e2xgrader.preprocessors import ValidateExtraCells

from ..test_utils.test_utils import create_temp_course


class TestClearSolutions(unittest.TestCase):
    def setUp(self):
        tmp_dir, coursedir = create_temp_course()
        self.nb = new_notebook()
        self.nb.cells = PresetManager(coursedir).get_question_preset("Single Choice")
        tmp_dir.cleanup()

    def test_invalid_notebook_with_extra_cells(self):
        with self.assertRaises(ValidationError):
            ValidateExtraCells().preprocess(self.nb, {})

    def test_valid_notebook_with_extra_cells(self):
        self.nb.cells[0].metadata.extended_cell.choice = ["1"]
        ValidateExtraCells().preprocess(self.nb, {})

    def test_valid_notebook_without_extra_cells(self):
        self.nb.cells = [new_markdown_cell()]
        ValidateExtraCells().preprocess(self.nb, {})



import glob
import hashlib
import os
import re

from nbgrader.exchange.default import ExchangeList
from nbgrader.utils import make_unique_key, notebook_hash

from .exchange import E2xExchange


def _checksum(path):
    m = hashlib.md5()
    m.update(open(path, "rb").read())
    return m.hexdigest()


def _get_key(info):
    return info["course_id"], info["student_id"], info["assignment_id"]


def _match_key(info, key):
    return (
        info["course_id"] == key[0]
        and info["student_id"] == key[1]
        and info["assignment_id"] == key[2]
    )


class E2xExchangeList(E2xExchange, ExchangeList):
    def init_dest(self):
        course_id = self.coursedir.course_id if self.coursedir.course_id else "*"
        assignment_id = (
            self.coursedir.assignment_id if self.coursedir.assignment_id else "*"
        )
        student_id = self.coursedir.student_id if self.coursedir.student_id else "*"

        if self.inbound:
            pattern = os.path.join(
                self.root,
                course_id,
                self.inbound_directory,
                "{}+{}+*".format(student_id, assignment_id),
            )
        elif self.cached:
            pattern = os.path.join(
                self.cache, course_id, "{}+{}+*".format(student_id, assignment_id)
            )
        else:
            if self.personalized_outbound:
                # list all assignments here
                pattern = os.path.join(
                    self.root,
                    course_id,
                    self.outbound_directory,
                    student_id,
                    "{}".format(assignment_id),
                )
            else:
                pattern = os.path.join(
                    self.root,
                    course_id,
                    self.outbound_directory,
                    "{}".format(assignment_id),
                )

        self.assignments = sorted(glob.glob(pattern))

    def parse_assignment(self, assignment):
        course_id = r".*/(?P<course_id>.*)/"
        if self.inbound:
            regexp = (
                course_id
                + self.inbound_directory
                + r"/(?P<student_id>[^+]*)\+"
                + r"(?P<assignment_id>[^+]*)\+"
                + r"(?P<timestamp>[^+]*)"
                + r"(?P<random_string>\+.*)?"
            )
        elif self.cached:
            regexp = (
                course_id
                + r"(?P<student_id>.*)\+"
                + r"(?P<assignment_id>.*)\+"
                + r"(?P<timestamp>.*)"
            )
        else:
            if self.personalized_outbound:
                regexp = (
                    course_id
                    + self.outbound_directory
                    + r"/(?P<student_id>.*)"
                    + r"/(?P<assignment_id>.*)"
                )
            else:
                regexp = course_id + self.outbound_directory + r"/(?P<assignment_id>.*)"

        m = re.match(regexp, assignment)
        if m is None:
            raise RuntimeError(
                "Could not match '%s' with regexp '%s'", assignment, regexp
            )

        return m.groupdict()

    def parse_assignments(self):
        if self.coursedir.student_id:
            courses = self.authenticator.get_student_courses(self.coursedir.student_id)
        else:
            courses = None

        assignments = []
        released_assignments = []
        for path in self.assignments:
            info = self.parse_assignment(path)
            # if grader and the assignment is already known as released assignment, skip looking
            if (
                self.personalized_outbound
                and self.grader
                and info["assignment_id"] in released_assignments
            ):
                self.log.debug(
                    "Grader role and personalized-outbound are enabled, "
                    "and the assignment is known to be released already"
                )
                continue

            if courses is not None and info["course_id"] not in courses:
                continue

            if self.path_includes_course:
                assignment_dir = os.path.join(
                    self.assignment_dir, info["course_id"], info["assignment_id"]
                )
            else:
                assignment_dir = os.path.join(
                    self.assignment_dir, info["assignment_id"]
                )

            if self.inbound or self.cached:
                info["status"] = "submitted"
                info["path"] = path
            elif os.path.exists(assignment_dir):
                info["status"] = "fetched"
                info["path"] = os.path.abspath(assignment_dir)
            else:
                info["status"] = "released"
                info["path"] = path
                # update released assignments
                if self.personalized_outbound and self.grader:
                    released_assignments.append(info["assignment_id"])

            if self.remove:
                info["status"] = "removed"

            notebooks = sorted(glob.glob(os.path.join(info["path"], "*.ipynb")))
            if not notebooks:
                self.log.warning("No notebooks found in {}".format(info["path"]))

            info["notebooks"] = []
            for notebook in notebooks:
                nb_info = {
                    "notebook_id": os.path.splitext(os.path.split(notebook)[1])[0],
                    "path": os.path.abspath(notebook),
                }
                if info["status"] != "submitted":
                    info["notebooks"].append(nb_info)
                    continue

                nb_info["has_local_feedback"] = False
                nb_info["has_exchange_feedback"] = False
                nb_info["local_feedback_path"] = None
                nb_info["feedback_updated"] = False

                # Check whether feedback has been fetched already.
                local_feedback_dir = os.path.join(
                    assignment_dir, "feedback", info["timestamp"]
                )
                local_feedback_path = os.path.join(
                    local_feedback_dir, "{0}.html".format(nb_info["notebook_id"])
                )
                has_local_feedback = os.path.isfile(local_feedback_path)
                if has_local_feedback:
                    local_feedback_checksum = _checksum(local_feedback_path)
                else:
                    local_feedback_checksum = None

                # Also look to see if there is feedback available to fetch.
                # and check whether personalized-feedback is enabled
                if self.personalized_feedback:
                    exchange_feedback_path = os.path.join(
                        self.root,
                        info["course_id"],
                        self.feedback_directory,
                        info["student_id"],
                        info["assignment_id"],
                        "{0}.html".format(nb_info["notebook_id"]),
                    )
                else:
                    unique_key = make_unique_key(
                        info["course_id"],
                        info["assignment_id"],
                        nb_info["notebook_id"],
                        info["student_id"],
                        info["timestamp"],
                    )
                    self.log.debug("Unique key is: {}".format(unique_key))
                    nb_hash = notebook_hash(notebook, unique_key)
                    exchange_feedback_path = os.path.join(
                        self.root,
                        info["course_id"],
                        self.feedback_directory,
                        "{0}.html".format(nb_hash),
                    )

                has_exchange_feedback = os.path.isfile(exchange_feedback_path)
                if not has_exchange_feedback:
                    # Try looking for legacy feedback.
                    nb_hash = notebook_hash(notebook)
                    exchange_feedback_path = os.path.join(
                        self.root,
                        info["course_id"],
                        "feedback",
                        "{0}.html".format(nb_hash),
                    )
                    has_exchange_feedback = os.path.isfile(exchange_feedback_path)
                if has_exchange_feedback:
                    exchange_feedback_checksum = _checksum(exchange_feedback_path)
                else:
                    exchange_feedback_checksum = None

                nb_info["has_local_feedback"] = has_local_feedback
                nb_info["has_exchange_feedback"] = has_exchange_feedback
                if has_local_feedback:
                    nb_info["local_feedback_path"] = local_feedback_path
                if has_local_feedback and has_exchange_feedback:
                    nb_info["feedback_updated"] = (
                        exchange_feedback_checksum != local_feedback_checksum
                    )
                info["notebooks"].append(nb_info)

            if info["status"] == "submitted":
                if info["notebooks"]:
                    # List feedback if there exists for one of the notebooks files in path
                    has_local_feedback = any(
                        [nb["has_local_feedback"] for nb in info["notebooks"]]
                    )
                    has_exchange_feedback = any(
                        [nb["has_exchange_feedback"] for nb in info["notebooks"]]
                    )
                    feedback_updated = any(
                        [nb["feedback_updated"] for nb in info["notebooks"]]
                    )
                else:
                    has_local_feedback = False
                    has_exchange_feedback = False
                    feedback_updated = False

                info["has_local_feedback"] = has_local_feedback
                info["has_exchange_feedback"] = has_exchange_feedback
                info["feedback_updated"] = feedback_updated
                if has_local_feedback:
                    info["local_feedback_path"] = os.path.join(
                        assignment_dir, "feedback", info["timestamp"]
                    )
                else:
                    info["local_feedback_path"] = None

            assignments.append(info)

        # partition the assignments into groups for course/student/assignment
        if self.inbound or self.cached:
            assignment_keys = sorted(
                list(set([_get_key(info) for info in assignments]))
            )
            assignment_submissions = []
            for key in assignment_keys:
                submissions = [x for x in assignments if _match_key(x, key)]
                submissions = sorted(submissions, key=lambda x: x["timestamp"])
                info = {
                    "course_id": key[0],
                    "student_id": key[1],
                    "assignment_id": key[2],
                    "status": submissions[0]["status"],
                    "submissions": submissions,
                }
                assignment_submissions.append(info)
            assignments = assignment_submissions

        return assignments



import json

from nbformat.notebooknode import NotebookNode
from nbgrader import utils
from nbgrader.api import MissingEntry
from nbgrader.preprocessors import SaveCells as NbgraderSaveCells

from ..utils.extra_cells import is_multiplechoice, is_singlechoice


class SaveCells(NbgraderSaveCells):
    def _create_source_cell(self, cell: NotebookNode) -> None:
        grade_id = cell.metadata.nbgrader["grade_id"]

        try:
            source_cell = self.gradebook.find_source_cell(
                grade_id, self.notebook_id, self.assignment_id
            ).to_dict()
            del source_cell["name"]
            del source_cell["notebook"]
            del source_cell["assignment"]
        except MissingEntry:
            source_cell = {}

        source = cell.source

        if is_singlechoice(cell) or is_multiplechoice(cell):
            source = json.dumps(cell.metadata.extended_cell)

        source_cell.update(
            {
                "cell_type": cell.cell_type,
                "locked": utils.is_locked(cell),
                "source": source,
                "checksum": cell.metadata.nbgrader.get("checksum", None),
            }
        )

        self.new_source_cells[grade_id] = source_cell



from nbconvert.filters.highlight import Highlight2HTML, _pygments_highlight


class Highlight2HTMLwithLineNumbers(Highlight2HTML):
    def __call__(self, source, language=None, metadata=None):
        """
        Return a syntax-highlighted version of the input source as html output.
        Same as Highlight2HTML but with inline line numbers

        Parameters
        ----------
        source : str
            source of the cell to highlight
        language : str
            language to highlight the syntax of
        metadata : NotebookNode cell metadata
            metadata of the cell to highlight
        """
        from pygments.formatters import HtmlFormatter

        if not language:
            language = self.pygments_lexer

        return _pygments_highlight(
            source if len(source) > 0 else " ",
            # needed to help post processors:
            HtmlFormatter(cssclass=" highlight hl-" + language, linenos="inline"),
            language,
            metadata,
        )



import unittest
from typing import List

from nbformat.notebooknode import NotebookNode
from nbformat.v4 import new_markdown_cell

from e2xgrader.graders import SingleChoiceGrader


class TestSingleChoiceGrader(unittest.TestCase):
    def create_singlechoice_cell(
        self,
        grade_id: str,
        student_choices: List[int],
        instructor_choices: List[int],
        points: int = 5,
    ) -> NotebookNode:
        cell = new_markdown_cell()

        cell.metadata = {
            "nbgrader": {
                "grade": True,
                "grade_id": grade_id,
                "locked": False,
                "points": points,
                "schema_version": 3,
                "solution": True,
                "task": False,
            },
            "extended_cell": {
                "type": "singlechoice",
                "choice": student_choices,
                "source": {"choice": instructor_choices},
            },
        }

        cell.source = """
        - correct answer
        - wrong answer
        """
        return cell

    def test_correct_answer(self):
        grader = SingleChoiceGrader()
        cell = self.create_singlechoice_cell("mycell", [0], [0], 10)
        points, max_points = grader.determine_grade(cell)
        assert points == max_points
        assert points == 10

    def test_incorrect_answer(self):
        grader = SingleChoiceGrader()
        cell = self.create_singlechoice_cell("mycell", [1], [0], 10)
        points, max_points = grader.determine_grade(cell)
        assert max_points == 10
        assert points == 0



import base64
import os
import sys
from stat import (
    S_IRGRP,
    S_IROTH,
    S_IRUSR,
    S_ISGID,
    S_IWGRP,
    S_IWOTH,
    S_IWUSR,
    S_IXGRP,
    S_IXOTH,
    S_IXUSR,
)

import nbformat
from nbgrader.exchange.default import ExchangeSubmit
from nbgrader.utils import check_mode, get_username

from .exchange import E2xExchange
from .utils import (
    append_hashcode,
    append_timestamp,
    compute_hashcode,
    generate_html,
    generate_student_info,
    truncate_hashcode,
)


class E2xExchangeSubmit(E2xExchange, ExchangeSubmit):
    def init_dest(self):
        if self.coursedir.course_id == "":
            self.fail("No course id specified. Re-run with --course flag.")
        if not self.authenticator.has_access(
            self.coursedir.student_id, self.coursedir.course_id
        ):
            self.fail("You do not have access to this course.")

        self.inbound_path = os.path.join(
            self.root, self.coursedir.course_id, self.inbound_directory
        )

        if self.personalized_inbound:
            self.inbound_path = os.path.join(
                self.inbound_path, os.getenv("JUPYTERHUB_USER")
            )

            if not os.path.isdir(self.inbound_path):
                self.log.warning(
                    "Inbound directory doesn't exist, creating {}".format(
                        self.inbound_path
                    )
                )
                # 0777 with set GID so student instructors can read students' submissions
                self.ensure_directory(
                    self.inbound_path,
                    S_ISGID
                    | S_IRUSR
                    | S_IWUSR
                    | S_IXUSR
                    | S_IRGRP
                    | S_IWGRP
                    | S_IXGRP
                    | S_IROTH
                    | S_IWOTH
                    | S_IXOTH
                    | (S_IRGRP if self.coursedir.groupshared else 0),
                )

        if not os.path.isdir(self.inbound_path):
            self.fail("Inbound directory doesn't exist: {}".format(self.inbound_path))
        if not check_mode(self.inbound_path, write=True, execute=True):
            self.fail(
                "You don't have write permissions to the directory: {}".format(
                    self.inbound_path
                )
            )

        self.cache_path = os.path.join(self.cache, self.coursedir.course_id)
        if self.coursedir.student_id != "*":
            # An explicit student id has been specified on the command line; we use it as student_id
            if "*" in self.coursedir.student_id or "+" in self.coursedir.student_id:
                self.fail(
                    "The student ID should contain no '*' nor '+'; got {}".format(
                        self.coursedir.student_id
                    )
                )
            student_id = self.coursedir.student_id
        else:
            student_id = get_username()
        if self.add_random_string:
            random_str = base64.urlsafe_b64encode(os.urandom(9)).decode("ascii")
            self.assignment_filename = "{}+{}+{}+{}".format(
                student_id, self.coursedir.assignment_id, self.timestamp, random_str
            )
        else:
            self.assignment_filename = "{}+{}+{}".format(
                student_id, self.coursedir.assignment_id, self.timestamp
            )

    def copy_files(self):
        self.init_release()

        hashcode = "No hashcode present"

        # Original notebook file
        student_notebook_file = os.path.join(
            self.src_path, f"{self.coursedir.assignment_id}.ipynb"
        )

        if os.path.isfile(student_notebook_file):
            nb = nbformat.read(student_notebook_file, as_version=nbformat.NO_CONVERT)
            nb = append_timestamp(nb, self.timestamp)
            nbformat.write(nb, student_notebook_file)
            hashcode = truncate_hashcode(
                compute_hashcode(student_notebook_file, method="sha1")
            )

            username = get_username()
            generate_student_info(
                os.path.join(self.src_path, f"{username}_info.txt"),
                username,
                hashcode,
                self.timestamp,
            )
            nb = append_hashcode(nb, hashcode)
            generate_html(
                nb,
                os.path.join(
                    self.src_path, f"{self.coursedir.assignment_id}_hashcode.html"
                ),
            )
        else:
            self.log.warning(
                "Can not generate hashcode, notebook and assignment name does not match."
            )

        dest_path = os.path.join(self.inbound_path, self.assignment_filename)
        if self.add_random_string:
            cache_path = os.path.join(
                self.cache_path, self.assignment_filename.rsplit("+", 1)[0]
            )
        else:
            cache_path = os.path.join(self.cache_path, self.assignment_filename)

        self.log.info("Source: {}".format(self.src_path))
        self.log.info("Destination: {}".format(dest_path))

        # copy to the real location
        self.check_filename_diff()
        self.do_copy(self.src_path, dest_path)
        with open(os.path.join(dest_path, "timestamp.txt"), "w") as fh:
            fh.write(self.timestamp)
        self.set_perms(
            dest_path,
            fileperms=(S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH),
            dirperms=(
                S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH
            ),
        )

        # Make this 0777=ugo=rwx so the instructor can delete later.
        # Hidden from other users by the timestamp.
        os.chmod(
            dest_path,
            S_IRUSR
            | S_IWUSR
            | S_IXUSR
            | S_IRGRP
            | S_IWGRP
            | S_IXGRP
            | S_IROTH
            | S_IWOTH
            | S_IXOTH,
        )

        # also copy to the cache
        if not os.path.isdir(self.cache_path):
            os.makedirs(self.cache_path)
        self.do_copy(self.src_path, cache_path)
        with open(os.path.join(cache_path, "timestamp.txt"), "w") as fh:
            fh.write(self.timestamp)

        self.log.info(
            "Submitted as: {} {} {}".format(
                self.coursedir.course_id,
                self.coursedir.assignment_id,
                str(self.timestamp),
            )
        )

        return hashcode, self.timestamp

    def init_release(self):
        if self.coursedir.course_id == "":
            self.fail("No course id specified. Re-run with --course flag.")

        course_path = os.path.join(self.root, self.coursedir.course_id)
        outbound_path = os.path.join(course_path, self.outbound_directory)
        if self.personalized_outbound:
            self.release_path = os.path.join(
                outbound_path,
                os.getenv("JUPYTERHUB_USER"),
                self.coursedir.assignment_id,
            )
        else:
            self.release_path = os.path.join(
                outbound_path, self.coursedir.assignment_id
            )

        if not os.path.isdir(self.release_path):
            self.fail("Assignment not found: {}".format(self.release_path))
        if not check_mode(self.release_path, read=True, execute=True):
            self.fail(
                "You don't have read permissions for the directory: {}".format(
                    self.release_path
                )
            )

    def start(self):
        if sys.platform == "win32":
            self.fail("Sorry, the exchange is not available on Windows.")
        if not self.coursedir.groupshared:
            # This just makes sure that directory is o+rwx.  In group shared
            # case, it is up to admins to ensure that instructors can write
            # there.
            self.ensure_root()

        self.set_timestamp()
        self.init_src()
        self.init_dest()
        return self.copy_files()



from .clearhiddentests import ClearHiddenTests
from .clearsolutions import ClearSolutions
from .extractattachments import ExtractAttachments
from .filtercellsbyid import FilterCellsById
from .filtertests import FilterTests
from .overwritecells import OverwriteCells
from .permutetasks import PermuteTasks
from .saveautogrades import SaveAutoGrades
from .savecells import SaveCells
from .scramble import Scramble
from .unpermutetasks import UnpermuteTasks
from .unscramble import Unscramble
from .validateextracells import ValidateExtraCells

__all__ = [
    "FilterCellsById",
    "SaveCells",
    "ClearSolutions",
    "ClearHiddenTests",
    "OverwriteCells",
    "SaveAutoGrades",
    "PermuteTasks",
    "UnpermuteTasks",
    "Scramble",
    "Unscramble",
    "ValidateExtraCells",
    "ExtractAttachments",
    "FilterTests",
]



from e2xcore import BaseApp

from .handlers import E2xAssignmentList, default_handlers


class AssignmentList(BaseApp):
    def load_app(self):
        self.log.info("Loading the e2x assignment list app")
        lister = E2xAssignmentList(parent=self.parent)
        lister.root_dir = self.webapp.settings["root_dir"]

        self.update_tornado_settings(dict(assignment_list_manager=lister))

        self.add_handlers(default_handlers)



from e2xauthoring.app import AuthoringApp
from traitlets import Any, List

from ..apps.assignment_list import AssignmentList
from ..apps.e2xgraderapi import E2xGraderApi
from ..apps.formgrader import FormgradeApp
from ..apps.help import Help
from ..apps.nbgraderapi import NbGraderApi
from ..apps.validate_assignment import ValidateAssignment
from ..base import BaseExtension


class TeacherExtension(BaseExtension):
    apps = List(
        trait=Any(),
        default_value=[
            E2xGraderApi,
            NbGraderApi,
            FormgradeApp,
            AuthoringApp,
            ValidateAssignment,
            AssignmentList,
            Help,
        ],
    ).tag(config=True)


def load_jupyter_server_extension(nbapp):
    """Load the e2xgrader serverextension"""
    nbapp.log.info("Loading the e2xgrader teacher serverextension")
    TeacherExtension(parent=nbapp)



from e2xcore.utils.nbgrader_cells import get_tasks, grade_id
from nbformat.v4 import new_notebook
from nbgrader.preprocessors import NbGraderPreprocessor


class UnpermuteTasks(NbGraderPreprocessor):
    def unpermute(self, nb):
        unpermuted_nb = new_notebook()
        # Get original order
        original_order = nb.metadata.original_order
        # Get indices of tasks
        tasks = get_tasks(nb)

        cursor = 0

        for task in tasks:
            # Add all cells between cursor and current task
            for idx in range(cursor, task[0]):
                unpermuted_nb.cells.append(nb.cells[idx])
            # Get name and cells of next task in the original order
            next_task_id = original_order.pop(0)
            next_task = [
                task for task in tasks if grade_id(nb.cells[task[0]]) == next_task_id
            ][0]
            # Add cells of the next task
            unpermuted_nb.cells.extend([nb.cells[idx] for idx in next_task])
            # Set the cursor to the idx after the inserted task
            cursor = task[-1] + 1

        # Add remaining cells at the bottom of the notebook
        unpermuted_nb.cells.extend(
            [nb.cells[idx] for idx in range(cursor, len(nb.cells))]
        )

        return unpermuted_nb

    def preprocess(self, nb, resources):
        if "original_order" not in nb.metadata:
            return nb, resources

        perm = nb.metadata.permutation
        keys = sorted(range(len(perm)), key=perm.__getitem__)
        reorder_cells = [nb.cells[i] for i in keys]
        nb.cells = reorder_cells
        return nb, resources



import os

from e2xcore import BaseApp
from nbgrader.apps.baseapp import NbGrader
from nbgrader.server_extensions.formgrader import handlers as nbgrader_handlers
from tornado import web

from e2xgrader.exporters import E2xExporter
from e2xgrader.preprocessors import FilterCellsById

from .handlers import default_handlers


class FormgradeApp(NbGrader, BaseApp):
    template_path = os.path.join(os.path.dirname(__file__), "templates")
    static_path = os.path.join(os.path.dirname(__file__), "static")

    def __init__(self, **kwargs):
        NbGrader.__init__(self, **kwargs)
        BaseApp.__init__(self, **kwargs)
        self.initialize([])
        if not self.config.has_key(
            "HTMLExporter"
        ) or not self.config.HTMLExporter.has_key("template_name"):
            self.config.HTMLExporter.template_name = "formgrade"

    def load_app(self):
        self.log.info("Loading the formgrader app")
        exporter = E2xExporter(config=self.config)
        exporter.register_preprocessor(FilterCellsById)

        self.update_tornado_settings(
            dict(
                nbgrader_jinja2_env=self.webapp.settings["e2xgrader"]["jinja_env"],
                nbgrader_exporter=exporter,
            )
        )
        self.add_template_path(self.template_path)
        self.add_template_path(nbgrader_handlers.template_path)
        self.add_handlers(default_handlers)
        self.add_handlers(nbgrader_handlers.default_handlers)

        static_handlers = [
            (
                r"/formgrader/static/(.*)",
                web.StaticFileHandler,
                {"path": nbgrader_handlers.static_path},
            ),
            (
                r"/e2xgrader/static/(.*)",
                web.StaticFileHandler,
                {"path": self.static_path},
            ),
        ]

        self.add_handlers(static_handlers)



from logging import Logger
from typing import Optional, Tuple

from nbformat.notebooknode import NotebookNode
from nbgrader.utils import get_partial_grade, is_solution

from .base import BaseGrader


class CodeGrader(BaseGrader):
    def extract_grade(self, text, log: Logger = None):
        delimiter_start = "### BEGIN GRADE"
        delimiter_end = "### END GRADE"
        inside = False

        points = None

        grade_lines = []

        for line in text.split("\n"):
            if line.startswith(delimiter_start):
                inside = True
            elif line.startswith(delimiter_end):
                inside = False
                break
            elif inside:
                grade_lines.append(line)

        if len(grade_lines) > 0 and not inside:
            # Try casting result to float
            try:
                points = float("".join(grade_lines))
            except ValueError:
                pass
        return points

    def determine_grade(
        self, cell: NotebookNode, log: Logger = None
    ) -> Tuple[Optional[float], float]:
        max_points = float(cell.metadata["nbgrader"]["points"])

        if not self.cell_changed(cell):
            return 0, max_points
        elif is_solution(cell):
            return None, max_points

        for output in cell.outputs:
            # option 1: error, return 0
            if (
                output.output_type == "error"
                or output.output_type == "stream"
                and output.name == "stderr"
            ):
                return 0, max_points
            if output.output_type == "execute_result":
                partial_grade = get_partial_grade(output, max_points, log)
                return partial_grade, max_points
            if output.output_type == "stream" and output.name == "stdout":
                partial_grade = self.extract_grade(output.text, log)
                if partial_grade is not None:
                    return partial_grade, max_points

        return max_points, max_points



import unittest
from copy import deepcopy

from e2xauthoring.managers import PresetManager
from nbformat.v4 import new_notebook

from e2xgrader.preprocessors import ExtractAttachments

from ..test_utils.test_utils import create_temp_course


class TestExtractAttachments(unittest.TestCase):
    def setUp(self) -> None:
        self.tmp_dir, self.coursedir = create_temp_course()
        self.resources = dict(
            metadata=dict(path=self.coursedir.root),
            outputs=dict(),
            output_files_dir=self.coursedir.root,
        )

    def tearDown(self) -> None:
        self.tmp_dir.cleanup()

    def testPreprocessStandardCell(self):
        nb = new_notebook()
        nb.cells = PresetManager(self.coursedir).get_question_preset("Single Choice")

        processed_nb = deepcopy(nb)
        processed_nb, _ = ExtractAttachments().preprocess(
            nb=processed_nb, resources=self.resources
        )

        self.assertDictEqual(nb, processed_nb)

    def testPreprocessDiagramCell(self):
        nb = new_notebook()
        nb.cells = PresetManager(self.coursedir).get_question_preset("Diagram")

        processed_nb = deepcopy(nb)
        processed_nb, _ = ExtractAttachments().preprocess(
            nb=processed_nb, resources=self.resources
        )

        attachment_string = "![diagram](attachment:diagram.png)"

        for processed_cell, cell in zip(processed_nb.cells, nb.cells):
            if attachment_string in cell.source:
                assert attachment_string not in processed_cell.source
                assert "<a href=" in processed_cell.source
                assert "diagram.png" in processed_cell.attachments



import os
import unittest

from notebook.serverextensions import BaseJSONConfigManager, jupyter_config_path

from e2xgrader.extensions import E2xExtensionManager


class TestE2XExtensionManager(unittest.TestCase):
    def setUp(self):
        self.serverextensions = [
            "nbgrader.server_extensions.formgrader",
            "nbgrader.server_extensions.validate_assignment",
            "nbgrader.server_extensions.assignment_list",
            "nbgrader.server_extensions.course_list",
            "e2xgrader.server_extensions.teacher",
            "e2xgrader.server_extensions.student",
        ]
        self.manager = E2xExtensionManager()
        self.manager.deactivate()

    def get_serverextensions(self, role):
        serverextensions = {
            "nbgrader.server_extensions.formgrader": [],
            "nbgrader.server_extensions.validate_assignment": [],
            "nbgrader.server_extensions.assignment_list": [],
            "nbgrader.server_extensions.course_list": [
                "teacher",
                "student",
                "student_exam",
            ],
            "e2xgrader.server_extensions.teacher": ["teacher"],
            "e2xgrader.server_extensions.student": [
                "student",
            ],
            "e2xgrader.server_extensions.student_exam": [
                "student_exam",
            ],
        }

        for key, value in serverextensions.items():
            serverextensions[key] = role in value

        return serverextensions

    def get_nbextensions(self, role):
        nbextensions = {
            "tree": {
                "teacher_tree/main": ["teacher"],
                "assignment_list/main": ["teacher", "student", "student_exam"],
                "student_exam_tree/main": ["student_exam"],
                "student_tree/main": ["student"],
            },
            "notebook": {
                "teacher_notebook/main": ["teacher"],
                "validate_assignment/main": ["teacher"],
                "student_notebook/main": ["student"],
                "student_exam_notebook/main": ["student_exam"],
            },
        }

        for section_dict in nbextensions.values():
            for key, value in section_dict.items():
                section_dict[key] = role in value

        return nbextensions

    def test_deactivated_serverextensions(self):
        config_dict = {}
        for config_dir in jupyter_config_path():
            cm = BaseJSONConfigManager(config_dir=config_dir)
            config_dict.update(cm.get("jupyter_notebook_config"))
        extensions = config_dict["NotebookApp"]["nbserver_extensions"]
        for serverextension in self.serverextensions:
            assert not extensions[serverextension]

    def test_teacher_serverextensions(self):
        teacher_serverextensions = self.get_serverextensions("teacher")
        self.manager.activate_teacher()

        config_dict = {}
        for config_dir in jupyter_config_path():
            cm = BaseJSONConfigManager(config_dir=config_dir)
            config_dict.update(cm.get("jupyter_notebook_config"))
        extensions = config_dict["NotebookApp"]["nbserver_extensions"]
        for serverextension, status in teacher_serverextensions.items():
            assert extensions[serverextension] == status

    def test_teacher_nbextensions(self):
        teacher_nbextensions = self.get_nbextensions("teacher")
        self.manager.activate_teacher()
        for section, extensions in teacher_nbextensions.items():
            config_dict = {}
            for config_dir in jupyter_config_path():
                config_dir = os.path.join(config_dir, "nbconfig")
                cm = BaseJSONConfigManager(config_dir=config_dir)
                config_dict.update(cm.get(section))

            for extension, status in extensions.items():
                if extension in config_dict["load_extensions"]:
                    assert config_dict["load_extensions"][extension] == status
                else:
                    assert not status

    def test_student_serverextensions(self):
        student_serverextensions = self.get_serverextensions("student")
        self.manager.activate_student()

        config_dict = {}
        for config_dir in jupyter_config_path():
            cm = BaseJSONConfigManager(config_dir=config_dir)
            config_dict.update(cm.get("jupyter_notebook_config"))
        extensions = config_dict["NotebookApp"]["nbserver_extensions"]
        for serverextension, status in student_serverextensions.items():
            assert extensions[serverextension] == status

    def test_student_nbextensions(self):
        student_nbextensions = self.get_nbextensions("student")
        self.manager.activate_student()
        for section, extensions in student_nbextensions.items():
            config_dict = {}
            for config_dir in jupyter_config_path():
                config_dir = os.path.join(config_dir, "nbconfig")
                cm = BaseJSONConfigManager(config_dir=config_dir)
                config_dict.update(cm.get(section))

            for extension, status in extensions.items():
                if extension in config_dict["load_extensions"]:
                    print("Testing extension", extension)
                    assert config_dict["load_extensions"][extension] == status
                else:
                    assert not status

    def test_student_exam_serverextensions(self):
        student_serverextensions = self.get_serverextensions("student_exam")
        self.manager.activate_student_exam()

        config_dict = {}
        for config_dir in jupyter_config_path():
            cm = BaseJSONConfigManager(config_dir=config_dir)
            config_dict.update(cm.get("jupyter_notebook_config"))
        extensions = config_dict["NotebookApp"]["nbserver_extensions"]
        for serverextension, status in student_serverextensions.items():
            assert extensions[serverextension] == status

    def test_student_exam_nbextensions(self):
        student_exam_nbextensions = self.get_nbextensions("student_exam")
        self.manager.activate_student_exam()
        for section, extensions in student_exam_nbextensions.items():
            config_dict = {}
            for config_dir in jupyter_config_path():
                config_dir = os.path.join(config_dir, "nbconfig")
                cm = BaseJSONConfigManager(config_dir=config_dir)
                config_dict.update(cm.get(section))

            for extension, status in extensions.items():
                if extension in config_dict["load_extensions"]:
                    assert config_dict["load_extensions"][extension] == status
                else:
                    assert not status

    def tearDown(self):
        self.manager.deactivate()



from logging import Logger
from typing import Optional, Tuple

from nbformat.notebooknode import NotebookNode

from ..utils.extra_cells import get_choices, get_instructor_choices, get_num_of_choices
from .base import BaseGrader


class MultipleChoiceGrader(BaseGrader):
    def determine_grade(
        self, cell: NotebookNode, log: Logger = None
    ) -> Tuple[Optional[float], float]:
        max_points = float(cell.metadata["nbgrader"]["points"])
        student_choices = get_choices(cell)
        instructor_choices = get_instructor_choices(cell)
        option_points = max_points / get_num_of_choices(cell)

        points = 0

        for i in range(get_num_of_choices(cell)):
            if ((i in student_choices) and (i in instructor_choices)) or (
                (i not in student_choices) and (i not in instructor_choices)
            ):
                points += option_points
            else:
                points -= option_points

        return max(0, round(points, 1)), max_points



import json
import os

from e2xcore import urljoin
from e2xcore.handlers import E2xApiHandler
from tornado import web


class ListFilesHandler(E2xApiHandler):
    def list_files(self, path_id, path):
        self.log.info(f"Listing files from {path}")
        if path is None or not os.path.exists(path):
            return []
        file_list = []
        exclude_dirs = []

        for root, dirs, files in os.walk(path):
            for name in dirs:
                if name.startswith("."):
                    exclude_dirs.append(name)
                elif os.path.isfile(os.path.join(root, name, "index.html")):
                    exclude_dirs.append(name)
                    file_list.append(
                        (
                            os.path.join(root, name),
                            os.path.join(root, name, "index.html"),
                        )
                    )

            dirs[:] = [d for d in dirs if d not in exclude_dirs]

            for name in files:
                file_list.append((os.path.join(root, name), os.path.join(root, name)))

        file_list = [
            (
                os.path.relpath(name, path),
                os.path.join(path_id, os.path.relpath(url, path)),
            )
            for name, url in file_list
        ]

        return file_list

    @web.authenticated
    def get(self):
        files = []
        for path_id, path in self.settings.get("e2xhelp_shared_dirs", dict()).items():
            files.extend(self.list_files(path_id, path))
        self.write(json.dumps(files))


api_url = urljoin("e2x", "help", "api")
default_handlers = [
    (urljoin(api_url, "files"), ListFilesHandler),
]



from .base import BaseExtensionManager
from .utils import get_serverextension_toggle

NBGRADER_ASSIGNMENT_LIST = "nbgrader.server_extensions.assignment_list"
NBGRADER_FORMGRADER = "nbgrader.server_extensions.formgrader"
NBGRADER_VALIDATE_ASSIGNMENT = "nbgrader.server_extensions.validate_assignment"
NBGRADER_COURSE_LIST = "nbgrader.server_extensions.course_list"
NBGRADER_EXTENSIONS = [
    NBGRADER_ASSIGNMENT_LIST,
    NBGRADER_COURSE_LIST,
    NBGRADER_FORMGRADER,
    NBGRADER_VALIDATE_ASSIGNMENT,
]

E2XGRADER_TEACHER = "e2xgrader.server_extensions.teacher"
E2XGRADER_STUDENT = "e2xgrader.server_extensions.student"
E2XGRADER_STUDENT_EXAM = "e2xgrader.server_extensions.student_exam"
E2XGRADER_EXTENSIONS = [E2XGRADER_TEACHER, E2XGRADER_STUDENT, E2XGRADER_STUDENT_EXAM]

toggle_server_extension_python = get_serverextension_toggle()


class ServerExtensionManager(BaseExtensionManager):
    def __init__(self):
        super().__init__()
        if toggle_server_extension_python is None:
            self.log.warn("No server extension toggler found!")

    def deactivate(self, sys_prefix: bool = True, user: bool = False) -> None:
        for import_name in E2XGRADER_EXTENSIONS + NBGRADER_EXTENSIONS:
            toggle_server_extension_python(
                import_name=import_name, enabled=False, sys_prefix=sys_prefix, user=user
            )

    def activate_teacher(self, sys_prefix: bool = True, user: bool = False) -> None:
        kwargs = dict(sys_prefix=sys_prefix, user=user)
        self.deactivate(**kwargs)
        for import_name in [E2XGRADER_TEACHER, NBGRADER_COURSE_LIST]:
            toggle_server_extension_python(
                import_name=import_name, enabled=True, **kwargs
            )

    def activate_student(self, sys_prefix: bool = True, user: bool = False) -> None:
        kwargs = dict(sys_prefix=sys_prefix, user=user)
        self.deactivate(**kwargs)
        for import_name in [E2XGRADER_STUDENT, NBGRADER_COURSE_LIST]:
            toggle_server_extension_python(
                import_name=import_name, enabled=True, **kwargs
            )

    def activate_student_exam(self, sys_prefix=True, user=False):
        kwargs = dict(sys_prefix=sys_prefix, user=user)
        self.deactivate(**kwargs)
        for import_name in [E2XGRADER_STUDENT_EXAM, NBGRADER_COURSE_LIST]:
            toggle_server_extension_python(
                import_name=import_name, enabled=True, **kwargs
            )



import os
import sys
from binascii import a2b_base64

from nbgrader.preprocessors import NbGraderPreprocessor
from traitlets import Set, Unicode

from ..utils.extra_cells import is_attachment_cell, is_diagram


class ExtractAttachments(NbGraderPreprocessor):
    subdirectory = Unicode("attachments").tag(config=True)

    output_filename_template = Unicode("attach_{cell_index}_{name}").tag(config=True)

    extract_output_types = Set(
        {"image/png", "image/jpeg", "image/svg+xml", "application/pdf"}
    ).tag(config=True)

    def preprocess(self, nb, resources):
        # Get files directory if it has been specified
        self.path = os.path.join(resources["metadata"]["path"], self.subdirectory)
        os.makedirs(self.path, exist_ok=True)

        for cell_index, cell in enumerate(nb.cells):
            cell, resources = self.preprocess_cell(cell, resources, cell_index)

        return nb, resources

    def preprocess_cell(self, cell, resources, cell_index):
        """
        Apply a transformation on each cell,

        Parameters
        ----------
        cell : NotebookNode cell
            Notebook cell being processed
        resources : dictionary
            Additional resources used in the conversion process.  Allows
            preprocessors to pass variables into the Jinja engine.
        cell_index : int
            Index of the cell being processed (see base.py)
        """

        if not (is_attachment_cell(cell) or is_diagram(cell)):
            return cell, resources

        if is_diagram(cell):
            # Remove the attachment line from source
            cell.source = cell.source.replace("![diagram](attachment:diagram.png)", "")

        # Get files directory if it has been specified
        output_files_dir = resources.get("output_files_dir", None)

        # Make sure outputs key exists
        if not isinstance(resources["outputs"], dict):
            resources["outputs"] = {}

        to_delete = []

        # Loop through all of the attachments in the cell
        for name, attach in cell.get("attachments", {}).items():
            for mime, data in attach.items():
                if mime not in self.extract_output_types:
                    continue

                # Binary files are base64-encoded, SVG is already XML
                if mime in {"image/png", "image/jpeg", "application/pdf"}:
                    # data is b64-encoded as text (str, unicode),
                    # we want the original bytes
                    data = a2b_base64(data)
                elif sys.platform == "win32":
                    data = data.replace("\n", "\r\n").encode("UTF-8")
                else:
                    data = data.encode("UTF-8")

                filename = self.output_filename_template.format(
                    cell_index=cell_index,
                    name=name,
                )
                link = os.path.join(self.subdirectory, filename)

                if output_files_dir is not None:
                    filename = os.path.join(self.path, filename)

                if name.endswith(".gif") and mime == "image/png":
                    filename = filename.replace(".gif", ".png")

                # Write the attachments to a file and add a link

                with open(filename, "wb") as f:
                    f.write(data)

                if "image" in mime:
                    cell.source += f'\n<a href="{link}"><img src="{link}"></a>\n'
                else:
                    cell.source += f'\n<a href="{link}">{name}</a>\n'
                if not is_diagram(cell):
                    to_delete.append(name)

                # now we need to change the cell source so that it links to the
                # filename instead of `attachment:`
                attach_str = "attachment:" + name
                if attach_str in cell.source:
                    cell.source = cell.source.replace(attach_str, filename)

        for name in to_delete:
            del cell.attachments[name]

        return cell, resources



import os

from e2xcore import BaseApp
from nbgrader.apps.baseapp import NbGrader

from .apihandlers import default_handlers


class E2xGraderApi(NbGrader, BaseApp):
    template_path = os.path.join(os.path.dirname(__file__), "templates")

    def __init__(self, **kwargs):
        NbGrader.__init__(self, **kwargs)
        BaseApp.__init__(self, **kwargs)

    def load_app(self):
        self.log.info("Loading the e2xgrader api app")
        self.add_template_path(self.template_path)
        self.add_handlers(default_handlers)



c = get_config()

c.Autograde.sanitize_preprocessors = [
    "nbgrader.preprocessors.ClearOutput",
    "nbgrader.preprocessors.DeduplicateIds",
    "nbgrader.preprocessors.OverwriteKernelspec",
    "e2xgrader.preprocessors.OverwriteCells",
    "nbgrader.preprocessors.CheckCellMetadata",
    "e2xgrader.preprocessors.UnpermuteTasks",
    "e2xgrader.preprocessors.Unscramble",
]

c.Autograde.autograde_preprocessors = [
    "nbgrader.preprocessors.Execute",
    "nbgrader.preprocessors.LimitOutput",
    "e2xgrader.preprocessors.SaveAutoGrades",
    "nbgrader.preprocessors.AssignLatePenalties",
    "nbgrader.preprocessors.CheckCellMetadata",
]

c.GenerateAssignment.preprocessors = [
    "e2xgrader.preprocessors.ValidateExtraCells",
    "nbgrader.preprocessors.IncludeHeaderFooter",
    "nbgrader.preprocessors.LockCells",
    "e2xgrader.preprocessors.ClearSolutions",
    "nbgrader.preprocessors.ClearOutput",
    "nbgrader.preprocessors.CheckCellMetadata",
    "nbgrader.preprocessors.ComputeChecksums",
    "e2xgrader.preprocessors.SaveCells",
    "e2xgrader.preprocessors.ClearHiddenTests",
    "nbgrader.preprocessors.ClearMarkScheme",
    "nbgrader.preprocessors.ComputeChecksums",
    "nbgrader.preprocessors.CheckCellMetadata",
]

# Set up the exchange

# c.ExchangeFactory.exchange = 'e2xgrader.exchange.E2xExchange'
# c.ExchangeFactory.submit = 'e2xgrader.exchange.E2xExchangeSubmit'
# c.ExchangeFactory.collect = 'e2xgrader.exchange.E2xExchangeCollect'
# c.ExchangeFactory.fetch_assignment = 'e2xgrader.exchange.E2xExchangeFetchAssignment'
# c.ExchangeFactory.release_assignment = 'e2xgrader.exchange.E2xExchangeReleaseAssignment'
# c.ExchangeFactory.list = 'e2xgrader.exchange.E2xExchangeList'

# c.Exchange.personalized_inbound = True
# c.Exchange.personalized_outbound = True



from typing import Tuple

from nbconvert.exporters.exporter import ResourcesDict
from nbformat.notebooknode import NotebookNode
from nbgrader.preprocessors import NbGraderPreprocessor
from nbgrader.utils import is_grade, is_solution
from traitlets import Bool, Unicode


class FilterTests(NbGraderPreprocessor):
    hide_cells = Bool(
        False, help="Hide test cells in the feedback generated for students."
    ).tag(config=True)

    test_case_stub = Unicode(
        "# This test case is hidden #",
        help="The message the test case source is replaced with",
    ).tag(config=True)

    def preprocess_cell(
        self, cell: NotebookNode, resources: ResourcesDict, cell_index: int
    ) -> Tuple[NotebookNode, ResourcesDict]:
        if self.hide_cells and is_grade(cell) and not is_solution(cell):
            cell.source = self.test_case_stub
        return cell, resources



from typing import Any, Dict, Union

from nbformat.notebooknode import NotebookNode
from nbformat.v4 import new_code_cell, new_markdown_cell


def nbgrader_metadata(**kwargs) -> Dict[str, Union[int, str, bool]]:
    """Create an nbgrader metadata dictionary
    Updates the metadata with any keyword arguments provided

    Returns:
        Dict[str, Union[int, str, bool]]: metadata
    """
    metadata = {
        "grade": False,
        "grade_id": "cell1",
        "locked": False,
        "schema_version": 3,
        "solution": False,
        "task": False,
    }
    metadata.update(kwargs)
    return metadata


def extra_cell_metadata(
    type: str, options: Dict[str, Any] = None, **kwargs
) -> Dict[str, Any]:
    """Create an extra cell metadata dictionary
    Updates the metadata with any keyword arguments provided

    Args:
        type (str): the type of the extra cell
        options (Dict[str, Any], optional): An options dictionary. Defaults to None.

    Returns:
        Dict[str, Any]: metadata
    """
    metadata = dict(type=type, options=options if options is not None else dict())
    metadata.update(kwargs)
    return metadata


def new_autograder_test_cell(
    source: str = "", points: int = 0, grade_id: str = "test_cell"
) -> NotebookNode:
    return new_code_cell(
        source=source,
        metadata=dict(
            nbgrader=nbgrader_metadata(grade=True, points=points, grade_id=grade_id)
        ),
    )


def new_manually_graded_code_cell(
    source: str = "", grade_id: str = "autograded_answer"
) -> NotebookNode:
    return new_code_cell(
        source=source,
        metadata=dict(
            nbgrader=nbgrader_metadata(grade=True, solution=True, grade_id=grade_id)
        ),
    )


def new_autograded_code_cell(
    source: str = "", points: int = 0, grade_id: str = "manual_answer"
) -> NotebookNode:
    return new_code_cell(
        source=source,
        metadata=dict(
            nbgrader=nbgrader_metadata(
                grade=False, solution=True, points=points, grade_id=grade_id
            )
        ),
    )


def new_manually_graded_markdown_cell(
    source: str = "", points: int = 0, grade_id: str = "manual_text_answer"
) -> NotebookNode:
    return new_markdown_cell(
        source=source,
        metadata=dict(
            nbgrader=nbgrader_metadata(
                grade=True, solution=True, points=points, grade_id=grade_id
            )
        ),
    )


def new_readonly_code_cell(
    source: str = "", grade_id: str = "code_example"
) -> NotebookNode:
    return new_code_cell(
        source=source,
        metadata=dict(nbgrader=nbgrader_metadata(locked=True, grade_id=grade_id)),
    )


def new_readonly_markdown_cell(
    source: str = "", grade_id: str = "description"
) -> NotebookNode:
    return new_markdown_cell(
        source=source,
        metadata=dict(nbgrader=nbgrader_metadata(locked=True, grade_id=grade_id)),
    )


def new_singlechoice_cell(
    source: str = "", points: int = 0, grade_id: str = "singlechoice", **kwargs
) -> NotebookNode:
    cell = new_manually_graded_markdown_cell(
        source=source, grade_id=grade_id, points=points
    )
    cell.metadata["extended_cell"] = extra_cell_metadata(type="singlechoice", **kwargs)
    return cell


def new_multiplechoice_cell(
    source: str = "", points: int = 0, grade_id: str = "multiplechoice", **kwargs
) -> NotebookNode:
    cell = new_manually_graded_markdown_cell(
        source=source, grade_id=grade_id, points=points
    )
    cell.metadata["extended_cell"] = extra_cell_metadata(
        type="multiplechoice", **kwargs
    )
    return cell


def new_diagram_cell(
    source: str = "",
    points: int = 0,
    grade_id: str = "diagram",
    diagram: str = None,
    **kwargs,
):
    attachments = dict()
    if diagram is not None:
        attachments = {"diagram.png": {"image/png": diagram}}
    return new_markdown_cell(
        source=source,
        attachments=attachments,
        metadata=dict(
            nbgrader=nbgrader_metadata(
                grade=True, solution=True, points=points, grade_id=grade_id
            ),
            extended_cell=extra_cell_metadata(type="diagram", **kwargs),
        ),
    )


def new_upload_cell(
    source: str = "",
    points: int = 0,
    grade_id: str = "attachments",
    attachments: Dict[str, Dict[str, str]] = None,
    **kwargs,
):
    return new_markdown_cell(
        source=source,
        attachments=attachments if attachments is not None else dict(),
        metadata=dict(
            nbgrader=nbgrader_metadata(
                grade=True, solution=True, points=points, grade_id=grade_id
            ),
            extended_cell=extra_cell_metadata(type="attachments", **kwargs),
        ),
    )



import random
from copy import deepcopy

from e2xcore.utils.nbgrader_cells import get_tasks, grade_id
from nbformat.v4 import new_notebook
from nbgrader.preprocessors import NbGraderPreprocessor


class PermuteTasks(NbGraderPreprocessor):
    def __init__(self, **kw):
        self.rand = random.Random()
        if kw and "seed" in kw:
            self.rand = random.Random(kw["seed"])

    def permute(self, nb):
        permuted_nb = new_notebook()
        # Get indices of tasks
        tasks = get_tasks(nb)
        # Shuffle tasks
        shuffled_tasks = deepcopy(tasks)
        random.shuffle(shuffled_tasks)
        # Record original order
        original_order = []

        cursor = 0

        for task in tasks:
            # Add all cells between cursor and current task
            for idx in range(cursor, task[0]):
                permuted_nb.cells.append(nb.cells[idx])
            # Get next task
            next_task = shuffled_tasks.pop(0)
            original_order.append(grade_id(nb.cells[task[0]]))
            # Add cells of the next task
            permuted_nb.cells.extend([nb.cells[idx] for idx in next_task])
            # Set the cursor to the idx after the inserted task
            cursor = task[-1] + 1

        # Add remaining cells at the bottom of the notebook
        permuted_nb.cells.extend(
            [nb.cells[idx] for idx in range(cursor, len(nb.cells))]
        )
        # Save order
        permuted_nb.metadata["original_order"] = original_order

        return permuted_nb

    def preprocess(self, nb, resources):
        return self.permute(nb), resources



from .collect import E2xExchangeCollect
from .exchange import E2xExchange
from .fetch_assignment import E2xExchangeFetchAssignment
from .fetch_feedback import E2xExchangeFetchFeedback
from .list import E2xExchangeList
from .release_assignment import E2xExchangeReleaseAssignment
from .release_feedback import E2xExchangeReleaseFeedback
from .submit import E2xExchangeSubmit

__all__ = [
    "E2xExchangeFetchAssignment",
    "E2xExchangeSubmit",
    "E2xExchangeCollect",
    "E2xExchangeReleaseAssignment",
    "E2xExchange",
    "E2xExchangeList",
    "E2xExchangeReleaseFeedback",
    "E2xExchangeFetchFeedback",
]



import os
import shutil
from stat import (
    S_IRGRP,
    S_IROTH,
    S_IRUSR,
    S_ISGID,
    S_IWGRP,
    S_IWOTH,
    S_IWUSR,
    S_IXGRP,
    S_IXOTH,
    S_IXUSR,
)

from nbgrader.exchange.default import ExchangeReleaseAssignment

from .exchange import E2xExchange


class E2xExchangeReleaseAssignment(E2xExchange, ExchangeReleaseAssignment):
    def init_dest(self):
        if self.coursedir.course_id == "":
            self.fail("No course id specified. Re-run with --course flag.")

        self.course_path = os.path.join(self.root, self.coursedir.course_id)
        self.outbound_path = os.path.join(self.course_path, self.outbound_directory)
        self.inbound_path = os.path.join(self.course_path, self.inbound_directory)

        # 0755
        # groupshared: +2040
        self.ensure_directory(
            self.course_path,
            S_IRUSR
            | S_IWUSR
            | S_IXUSR
            | S_IRGRP
            | S_IXGRP
            | S_IROTH
            | S_IXOTH
            | ((S_ISGID | S_IWGRP) if self.coursedir.groupshared else 0),
        )

        if self.personalized_outbound:
            # the dest path is <exchange_root>/personalized-outbound/<username>/<assignment_id>
            # <username> is created by either grader via formgrader or student during spawning
            # <username> here is deterministic and has to be in the release directory
            self.dest_path = self.outbound_path

            # 0777 for personalized_outbound
            # groupshared: +2040
            self.ensure_directory(
                self.dest_path,
                S_IRUSR
                | S_IWUSR
                | S_IXUSR
                | S_IRGRP
                | S_IWGRP
                | S_IXGRP
                | S_IROTH
                | S_IWOTH
                | S_IXOTH
                | ((S_ISGID | S_IWGRP) if self.coursedir.groupshared else 0),
            )
        else:
            self.dest_path = os.path.join(
                self.outbound_path, self.coursedir.assignment_id
            )

            # 0755
            # groupshared: +2040
            self.ensure_directory(
                self.outbound_path,
                S_IRUSR
                | S_IWUSR
                | S_IXUSR
                | S_IRGRP
                | S_IXGRP
                | S_IROTH
                | S_IXOTH
                | ((S_ISGID | S_IWGRP) if self.coursedir.groupshared else 0),
            )

        # 0733 with set GID so student submission will have the instructors group
        # groupshared: +0040
        self.ensure_directory(
            self.inbound_path,
            S_ISGID
            | S_IRUSR
            | S_IWUSR
            | S_IXUSR
            | S_IWGRP
            | S_IXGRP
            | S_IWOTH
            | S_IXOTH
            | (S_IRGRP if self.coursedir.groupshared else 0),
        )

    def copy_files(self):
        if self.personalized_outbound:
            # check available users under course_path/<assignmet_dir>/<username>
            user_list = [
                user
                for user in os.listdir(self.src_path)
                if os.path.isdir(os.path.join(self.src_path, user))
                and user != self.coursedir.assignment_id
            ]
            for user in user_list:
                released_assignment_root = os.path.join(self.dest_path, user)
                released_user_assignment = os.path.join(
                    released_assignment_root, self.coursedir.assignment_id
                )
                if os.path.isdir(released_user_assignment):
                    shutil.rmtree(released_user_assignment)

                src_assignment = os.path.join(self.src_path, user)
                if os.path.isdir(src_assignment):
                    self.log.info(f"Source: {src_assignment}")
                    self.log.info(f"Destination: {released_user_assignment}")
                    self.do_copy(src_assignment, released_user_assignment)
                    self.set_released_assignment_perm(released_user_assignment)
                else:
                    self.log.info(f"Src assignment not found: {src_assignment}")
        else:
            if os.path.isdir(self.dest_path):
                if self.force:
                    self.log.info(
                        "Overwriting files: {} {}".format(
                            self.coursedir.course_id, self.coursedir.assignment_id
                        )
                    )
                    shutil.rmtree(self.dest_path)
                else:
                    self.fail(
                        "Destination already exists, add --force to overwrite: {} {}".format(
                            self.coursedir.course_id, self.coursedir.assignment_id
                        )
                    )

            self.log.info(f"Source: {self.src_path}")
            self.log.info(f"Destination: {self.dest_path}")
            self.do_copy(self.src_path, self.dest_path)
            self.set_released_assignment_perm(self.dest_path)
            self.log.info(
                "Released as: {} {}".format(
                    self.coursedir.course_id, self.coursedir.assignment_id
                )
            )

    def set_released_assignment_perm(self, path):
        self.set_perms(
            path,
            fileperms=(
                S_IRUSR
                | S_IWUSR
                | S_IRGRP
                | S_IROTH
                | (S_IWGRP if self.coursedir.groupshared else 0)
            ),
            dirperms=(
                S_IRUSR
                | S_IWUSR
                | S_IXUSR
                | S_IRGRP
                | S_IXGRP
                | S_IROTH
                | S_IXOTH
                | ((S_ISGID | S_IWGRP) if self.coursedir.groupshared else 0)
            ),
        )



import json
from typing import Tuple

from nbconvert.exporters.exporter import ResourcesDict
from nbformat.notebooknode import NotebookNode
from nbgrader.api import MissingEntry
from nbgrader.preprocessors import OverwriteCells as NbgraderOverwriteCells

from ..utils.extra_cells import is_multiplechoice, is_singlechoice


class OverwriteCells(NbgraderOverwriteCells):
    def preprocess_cell(
        self, cell: NotebookNode, resources: ResourcesDict, cell_index: int
    ) -> Tuple[NotebookNode, ResourcesDict]:
        if not (is_singlechoice(cell) or is_multiplechoice(cell)):
            return super().preprocess_cell(cell, resources, cell_index)

        grade_id = cell.metadata.get("nbgrader", {}).get("grade_id", None)
        if grade_id is None:
            return cell, resources
        try:
            source_cell = self.gradebook.find_source_cell(
                grade_id, self.notebook_id, self.assignment_id
            )
        except MissingEntry:
            self.log.warning(f"Cell {grade_id} does not exist in database")
            del cell.metadata.nbgrader["grade_id"]
            return cell, resources

        cell.metadata.extended_cell.source = json.loads(source_cell.source)

        return cell, resources



import traceback

from nbgrader.nbgraderformat import ValidationError
from nbgrader.preprocessors import NbGraderPreprocessor

from ..utils.extra_cells import is_singlechoice


class ExtraCellValidator:
    def validate_cell(self, cell):
        if "nbgrader" not in cell.metadata:
            return

        # check if there is a single choice cell without a solution
        if is_singlechoice(cell):
            extended_metadata = cell.metadata.extended_cell
            if ("choice" not in extended_metadata) or (
                len(extended_metadata.choice) < 1
            ):
                raise ValidationError(
                    "single choice nbgrader cell {} does not have a solution".format(
                        cell.metadata.nbgrader.grade_id
                    )
                )

    def validate_nb(self, nb):
        for cell in nb.cells:
            self.validate_cell(cell)


class ValidateExtraCells(NbGraderPreprocessor):
    """A preprocessor for checking that choice cells have valid solutions."""

    def preprocess(self, nb, resources):
        try:
            ExtraCellValidator().validate_nb(nb)
        except ValidationError:
            self.log.error(traceback.format_exc())
            msg = "Some choice cells seem to miss a solution. Please check them again."
            self.log.error(msg)
            raise ValidationError(msg)

        return nb, resources



from importlib import import_module
from types import ModuleType
from typing import Dict, List, Optional

from .. import _jupyter_nbextension_paths


def is_installed(package: str) -> bool:
    try:
        import_module(package)
        return True
    except ModuleNotFoundError:
        return False


def get_notebook_major_version() -> int:
    from notebook import __version__

    return int(__version__.split(".")[0])


def get_nbextension_utils() -> Optional[ModuleType]:
    if is_installed("notebook") and get_notebook_major_version() < 7:
        return import_module("notebook.nbextensions")
    if is_installed("nbclassic"):
        return import_module("nbclassic.nbextensions")
    return None


def discover_nbextensions(mode: str) -> List[Dict[str, str]]:
    extensions = list()
    for nbextension in _jupyter_nbextension_paths():
        if (
            f"{mode}_notebook" in nbextension["dest"]
            or f"{mode}_tree" in nbextension["dest"]
        ):
            extensions.append(
                dict(require=nbextension["require"], section=nbextension["section"])
            )
    return extensions


def get_serverextension_toggle() -> Optional[ModuleType]:
    def toggle_server_extension_python(**kwargs):
        if is_installed("notebook") and get_notebook_major_version() < 7:
            module = import_module("notebook.serverextensions")
            module.toggle_serverextension_python(**kwargs)
        if is_installed("jupyter_server"):
            module = import_module("jupyter_server.extension.serverextension")
            module.toggle_server_extension_python(**kwargs)

    return toggle_server_extension_python



from abc import abstractmethod

from traitlets.config import LoggingConfigurable


class BaseExtensionManager(LoggingConfigurable):
    @abstractmethod
    def deactivate(self, sys_prefix: bool = True, user: bool = False) -> None:
        pass

    @abstractmethod
    def activate_teacher(self, sys_prefix: bool = True, user: bool = False) -> None:
        pass

    @abstractmethod
    def activate_student(self, sys_prefix: bool = True, user: bool = False) -> None:
        pass

    @abstractmethod
    def activate_student_exam(self, sys_prefix=True, user=False):
        pass



from typing import Tuple

from nbconvert.exporters.exporter import ResourcesDict
from nbformat.notebooknode import NotebookNode
from nbgrader.preprocessors import ClearSolutions as NbgraderClearSolutions

from ..utils.extra_cells import get_options, is_diagram, is_extra_cell


class ClearSolutions(NbgraderClearSolutions):
    def preprocess_cell(
        self, cell: NotebookNode, resources: ResourcesDict, cell_index: int
    ) -> Tuple[NotebookNode, ResourcesDict]:
        if is_extra_cell(cell):
            if is_diagram(cell):
                # Check whether we want to remove the diagram
                options = get_options(cell)
                if options.get("replace_diagram", True):
                    del cell.attachments["diagram.png"]

            return cell, resources
        else:
            return super().preprocess_cell(cell, resources, cell_index)



import os

from e2xcore import format_url
from nbgrader.api import MissingEntry
from nbgrader.server_extensions.formgrader.base import (
    BaseHandler,
    check_notebook_dir,
    check_xsrf,
)
from nbgrader.server_extensions.formgrader.handlers import (
    SubmissionNavigationHandler as NbgraderSubmissionNavigationHandler,
)
from tornado import web


class ExportGradesHandler(BaseHandler):
    @web.authenticated
    @check_xsrf
    def get(self):
        html = self.render("export_grades.tpl", base_url=self.base_url)
        self.write(html)


class GradebookAssignmentsHandler(BaseHandler):
    @web.authenticated
    @check_xsrf
    @check_notebook_dir
    def get(self):
        view = self.get_argument("view", "notebook")
        if view == "task":
            template = os.path.join("task_view", "gradebook_assignments.tpl")
        else:
            template = "gradebook_assignments.tpl"
        html = self.render(template, base_url=self.base_url)
        self.write(html)


class GradebookNotebooksHandler(BaseHandler):
    @web.authenticated
    @check_xsrf
    @check_notebook_dir
    def get(self, assignment_id):
        view = self.get_argument("view", "notebook")
        if view == "task":
            template = os.path.join("task_view", "gradebook_notebooks.tpl")
        else:
            template = "gradebook_notebooks.tpl"
        html = self.render(
            template, assignment_id=assignment_id, base_url=self.base_url
        )
        self.write(html)


class GradebookTasksHandler(BaseHandler):
    @web.authenticated
    @check_xsrf
    def get(self, assignment_id, notebook_id):
        html = self.render(
            "task_view/gradebook_tasks.tpl",
            assignment_id=assignment_id,
            notebook_id=notebook_id,
            base_url=self.base_url,
        )
        self.write(html)


class GradebookNotebookSubmissionsHandler(BaseHandler):
    @web.authenticated
    @check_xsrf
    @check_notebook_dir
    def get(self, assignment_id, notebook_id):
        view = self.get_argument("view", "notebook")
        if view == "task":
            template = os.path.join("task_view", "gradebook_notebook_submissions.tpl")
        else:
            template = "gradebook_notebook_submissions.tpl"
        task_id = self.get_argument("filter", "")

        html = self.render(
            template,
            assignment_id=assignment_id,
            notebook_id=notebook_id,
            base_url=self.base_url,
            task_id=task_id,
        )
        self.write(html)


class SubmissionNavigationHandler(NbgraderSubmissionNavigationHandler):
    def _assignment_notebook_list_url(self, assignment_id, notebook_id, task_id):
        if len(task_id) < 1:
            return "{}/formgrader/gradebook/{}/{}".format(
                self.base_url, assignment_id, notebook_id
            )
        return format_url(
            f"{self.base_url}/formgrader/gradebook/{assignment_id}/{notebook_id}/",
            dict(view="task", filter=task_id),
        )

    def _next(self, assignment_id, notebook_id, submission, task_id):
        # find next submission
        submission_ids = self._get_submission_ids(assignment_id, notebook_id)
        ix = submission_ids.index(submission.id)
        if ix == (len(submission_ids) - 1):
            return self._assignment_notebook_list_url(
                assignment_id, notebook_id, task_id
            )
        else:
            return format_url(
                self._submission_url(submission_ids[ix + 1]), dict(task=task_id)
            )

    def _prev(self, assignment_id, notebook_id, submission, task_id):
        # find previous submission
        submission_ids = self._get_submission_ids(assignment_id, notebook_id)
        ix = submission_ids.index(submission.id)
        if ix == 0:
            return self._assignment_notebook_list_url(
                assignment_id, notebook_id, task_id
            )
        else:
            return format_url(
                self._submission_url(submission_ids[ix - 1]), dict(task=task_id)
            )

    def _next_incorrect(self, assignment_id, notebook_id, submission, task_id):
        # find next incorrect submission
        submission_ids = self._get_incorrect_submission_ids(
            assignment_id, notebook_id, submission
        )
        ix_incorrect = submission_ids.index(submission.id)
        if ix_incorrect == (len(submission_ids) - 1):
            return self._assignment_notebook_list_url(
                assignment_id, notebook_id, task_id
            )
        else:
            return format_url(
                self._submission_url(submission_ids[ix_incorrect + 1]),
                dict(task=task_id),
            )

    def _prev_incorrect(self, assignment_id, notebook_id, submission, task_id):
        # find previous incorrect submission
        submission_ids = self._get_incorrect_submission_ids(
            assignment_id, notebook_id, submission
        )
        ix_incorrect = submission_ids.index(submission.id)
        if ix_incorrect == 0:
            return self._assignment_notebook_list_url(
                assignment_id, notebook_id, task_id
            )
        else:
            return format_url(
                self._submission_url(submission_ids[ix_incorrect - 1]),
                dict(task=task_id),
            )

    @web.authenticated
    @check_xsrf
    @check_notebook_dir
    def get(self, submission_id, action):
        try:
            submission = self.gradebook.find_submission_notebook_by_id(submission_id)
            assignment_id = submission.assignment.assignment.name
            notebook_id = submission.notebook.name
            task_id = self.get_argument("task", "")
        except MissingEntry:
            raise web.HTTPError(404, "Invalid submission: {}".format(submission_id))

        handler = getattr(self, "_{}".format(action))
        self.redirect(
            handler(assignment_id, notebook_id, submission, task_id), permanent=False
        )


class SubmissionHandler(BaseHandler):
    @web.authenticated
    @check_xsrf
    @check_notebook_dir
    def get(self, submission_id):
        try:
            submission = self.gradebook.find_submission_notebook_by_id(submission_id)
            assignment_id = submission.assignment.assignment.name
            notebook_id = submission.notebook.name
            student_id = submission.student.id
        except MissingEntry:
            raise web.HTTPError(404, "Invalid submission: {}".format(submission_id))

        # redirect if there isn't a trailing slash in the uri
        if os.path.split(self.request.path)[1] == submission_id:
            url = self.request.path + "/"
            if self.request.query:
                url += "?" + self.request.query
            return self.redirect(url, permanent=True)

        filename = os.path.join(
            os.path.abspath(
                self.coursedir.format_path(
                    self.coursedir.autograded_directory, student_id, assignment_id
                )
            ),
            "{}.ipynb".format(notebook_id),
        )
        relative_path = os.path.relpath(filename, self.coursedir.root)
        indices = self.api.get_notebook_submission_indices(assignment_id, notebook_id)
        ix = indices.get(submission.id, -2)

        task_id = self.get_argument("task", "")

        resources = {
            "assignment_id": assignment_id,
            "notebook_id": notebook_id,
            "submission_id": submission.id,
            "index": ix,
            "total": len(indices),
            "base_url": self.base_url,
            "my_mathjax_url": self.base_url + "/" + self.mathjax_url,
            "student": student_id,
            "last_name": submission.student.last_name,
            "first_name": submission.student.first_name,
            "notebook_path": self.url_prefix + "/" + relative_path,
            "keyword": task_id,
        }

        if not os.path.exists(filename):
            resources["filename"] = filename
            html = self.render("formgrade_404.tpl", resources=resources)
            self.clear()
            self.set_status(404)
            self.write(html)

        else:
            html, _ = self.exporter.from_filename(filename, resources=resources)
            self.write(html)


root_path = os.path.dirname(__file__)
template_path = os.path.join(root_path, "templates")
static_path = os.path.join(root_path, "static")

_navigation_regex = r"(?P<action>next_incorrect|prev_incorrect|next|prev)"

default_handlers = [
    (r"/formgrader/export_grades/?", ExportGradesHandler),
    (r"/formgrader/gradebook/?", GradebookAssignmentsHandler),
    (r"/formgrader/gradebook/([^/]+)/?", GradebookNotebooksHandler),
    (r"/formgrader/gradebook/tasks/([^/]+)/([^/]+)/?", GradebookTasksHandler),
    (r"/formgrader/gradebook/([^/]+)/([^/]+)/?", GradebookNotebookSubmissionsHandler),
    (
        r"/formgrader/submissions/(?P<submission_id>[^/]+)/%s/?" % _navigation_regex,
        SubmissionNavigationHandler,
    ),
    (r"/formgrader/submissions/([^/]+)/?", SubmissionHandler),
]



from textwrap import dedent
from typing import Tuple

from e2xcore.utils.e2xgrader_cells import is_extra_cell
from e2xcore.utils.nbgrader_cells import grade_id
from nbconvert.exporters.exporter import ResourcesDict
from nbformat.notebooknode import NotebookNode
from nbgrader.api import Gradebook
from nbgrader.preprocessors import SaveAutoGrades as NbgraderSaveAutoGrades
from nbgrader.utils import determine_grade
from traitlets import Dict, Instance, List, TraitError, Unicode, validate

from ..graders import BaseGrader, CodeGrader, MultipleChoiceGrader, SingleChoiceGrader


class SaveAutoGrades(NbgraderSaveAutoGrades):
    graders = Dict(
        key_trait=Unicode(),
        value_trait=Instance(klass=BaseGrader),
        default_value={
            "code": CodeGrader(),
            "singlechoice": SingleChoiceGrader(),
            "multiplechoice": MultipleChoiceGrader(),
        },
    ).tag(config=True)

    cells = List(
        [],
        help=dedent(
            """
            List of cells to save the autogrades for. If this is empty all cells will be saved.
            """
        ),
    ).tag(config=True)

    @validate("cells")
    def _validate_cells(self, proposal):
        value = proposal["value"]
        if len(value) == 1:
            elem = value[0].strip()
            if elem.startswith("[") and elem.endswith("]"):
                elem = elem[1:-1]
            value = [v.strip() for v in elem.split(",")]
        if not isinstance(value, list):
            raise TraitError("cells must be a list")
        return value

    def cell_type(self, cell: NotebookNode):
        if is_extra_cell(cell):
            return cell.metadata.extended_cell.type
        return cell.cell_type

    def _add_score(self, cell: NotebookNode, resources: ResourcesDict) -> None:
        """Graders can override the autograder grades, and may need to
        manually grade written solutions anyway. This function adds
        score information to the database if it doesn't exist. It does
        NOT override the 'score' field, as this is the manual score
        that might have been provided by a grader.

        """
        # these are the fields by which we will identify the score
        # information
        grade = self.gradebook.find_grade(
            cell.metadata["nbgrader"]["grade_id"],
            self.notebook_id,
            self.assignment_id,
            self.student_id,
        )

        # determine what the grade is
        if self.cell_type(cell) in self.graders:
            auto_score, _ = self.graders[self.cell_type(cell)].determine_grade(
                cell, self.log
            )
            grade.auto_score = auto_score
        else:
            auto_score, _ = determine_grade(cell, self.log)
            grade.auto_score = auto_score

        # if there was previously a manual grade, or if there is no autograder
        # score, then we should mark this as needing review
        if (grade.manual_score is not None) or (grade.auto_score is None):
            grade.needs_manual_grade = True
        else:
            grade.needs_manual_grade = False

        self.gradebook.db.commit()

    def preprocess(
        self, nb: NotebookNode, resources: ResourcesDict
    ) -> Tuple[NotebookNode, ResourcesDict]:
        # pull information from the resources
        self.notebook_id = resources["nbgrader"]["notebook"]
        self.assignment_id = resources["nbgrader"]["assignment"]
        self.student_id = resources["nbgrader"]["student"]
        self.db_url = resources["nbgrader"]["db_url"]

        # connect to the database
        self.gradebook = Gradebook(self.db_url)

        with self.gradebook:
            # process the cells
            nb, resources = super(SaveAutoGrades, self).preprocess(nb, resources)

        return nb, resources

    def preprocess_cell(
        self, cell: NotebookNode, resources: ResourcesDict, cell_index: int
    ) -> Tuple[NotebookNode, ResourcesDict]:
        if len(self.cells) == 0 or grade_id(cell) in self.cells:
            return super().preprocess_cell(cell, resources, cell_index)
        return cell, resources



import unittest

from e2xgrader.preprocessors import ClearHiddenTests
from e2xgrader.tests.test_utils.cells import (
    new_autograder_test_cell,
    new_multiplechoice_cell,
)
from e2xgrader.utils.extra_cells import get_choices


class TestClearHiddenTests(unittest.TestCase):
    def test_choice_cell(self):
        source = """### Multiplechoice Question
        - Choice 1
        - Choice 2
        - Choice 3"""
        cell = new_multiplechoice_cell(
            source=source,
            choice=["1", "2"],
            num_of_choices=3,
        )

        assert len(get_choices(cell)) == 2
        assert all([i in get_choices(cell) for i in [1, 2]])

        processed_cell, _ = ClearHiddenTests().preprocess_cell(cell, {}, 0)
        assert len(get_choices(processed_cell)) == 0
        assert processed_cell.source == source

    def test_code_cell(self):
        cell = new_autograder_test_cell(
            source="""### BEGIN HIDDEN TESTS
        assert 1 == 1
        ###END HIDDEN TESTS"""
        )

        processed_cell, _ = ClearHiddenTests().preprocess_cell(cell, {}, 0)
        assert processed_cell.source == ""



import unittest

from e2xgrader.preprocessors import FilterTests

from ..test_utils.cells import (
    new_autograded_code_cell,
    new_autograder_test_cell,
    new_manually_graded_code_cell,
    new_manually_graded_markdown_cell,
    new_readonly_code_cell,
    new_readonly_markdown_cell,
)


class TestFilterTests(unittest.TestCase):
    def setUp(self) -> None:
        self.source = "assert 2**2 == 4"

    def test_hide_tests(self):
        preprocessor = FilterTests()
        preprocessor.hide_cells = True
        cell, _ = preprocessor.preprocess_cell(
            new_autograder_test_cell(source=self.source), {}, 0
        )
        assert cell.source == preprocessor.test_case_stub

    def test_show_test(self):
        preprocessor = FilterTests()
        preprocessor.hide_cells = False
        cell, _ = preprocessor.preprocess_cell(
            new_autograder_test_cell(source=self.source), {}, 0
        )
        assert cell.source == self.source

    def test_custom_test_stub(self):
        preprocessor = FilterTests()
        preprocessor.hide_cells = True
        custom_test_stub = "My custom message"
        preprocessor.test_case_stub = custom_test_stub
        cell, _ = preprocessor.preprocess_cell(
            new_autograder_test_cell(source=self.source), {}, 0
        )
        assert cell.source == custom_test_stub

    def test_ignore_non_test_cells(self):
        preprocessor = FilterTests()
        preprocessor.hide_cells = True

        cell, _ = preprocessor.preprocess_cell(
            new_manually_graded_code_cell(source=self.source), {}, 0
        )
        assert cell.source == self.source

        cell, _ = preprocessor.preprocess_cell(
            new_autograded_code_cell(source=self.source), {}, 0
        )
        assert cell.source == self.source

        cell, _ = preprocessor.preprocess_cell(
            new_manually_graded_markdown_cell(source=self.source), {}, 0
        )
        assert cell.source == self.source

        cell, _ = preprocessor.preprocess_cell(
            new_readonly_code_cell(source=self.source), {}, 0
        )
        assert cell.source == self.source

        cell, _ = preprocessor.preprocess_cell(
            new_readonly_markdown_cell(source=self.source), {}, 0
        )
        assert cell.source == self.source



import os

from e2xcore import BaseApp
from nbgrader.apps.baseapp import NbGrader
from nbgrader.server_extensions.formgrader.apihandlers import default_handlers


class NbGraderApi(NbGrader, BaseApp):
    def __init__(self, **kwargs):
        NbGrader.__init__(self, **kwargs)
        BaseApp.__init__(self, **kwargs)
        if self.parent.name == "jupyter-notebook":
            self.root_dir = self.parent.notebook_dir
        else:
            self.root_dir = self.parent.root_dir

    @property
    def url_prefix(self):
        relpath = os.path.relpath(self.coursedir.root, self.root_dir)
        return relpath

    def load_app(self):
        self.log.info("Loading the nbgrader api app")
        self.initialize([])

        course_dir = self.coursedir.root
        notebook_dir = self.parent.notebook_dir
        relpath = os.path.relpath(course_dir, notebook_dir)
        if relpath.startswith("../"):
            nbgrader_bad_setup = True
            self.log.error(
                "The course directory root is not a subdirectory of the notebook "
                "server root. This means that nbgrader will not work correctly. "
                "If you want to use nbgrader, please ensure the course directory "
                "root is in a subdirectory of the notebook root: %s",
                notebook_dir,
            )
        else:
            nbgrader_bad_setup = False

        # Save which kind of application is running : Jupyterlab like or classic Notebook
        self.webapp.settings["is_jlab"] = self.name != "jupyter-notebook"

        tornado_settings = dict(
            nbgrader_url_prefix=self.url_prefix,
            nbgrader_formgrader=self,
            nbgrader_coursedir=self.coursedir,
            nbgrader_authenticator=self.authenticator,
            nbgrader_gradebook=None,
            nbgrader_db_url=self.coursedir.db_url,
            nbgrader_bad_setup=nbgrader_bad_setup,
        )

        self.update_tornado_settings(tornado_settings)
        self.add_handlers(default_handlers)



import contextlib
import json
import os
import traceback

from nbgrader.auth import Authenticator
from nbgrader.coursedir import CourseDirectory
from nbgrader.exchange import ExchangeFactory
from nbgrader.server_extensions.assignment_list.handlers import (
    AssignmentList,
    AssignmentListHandler,
    BaseAssignmentHandler,
    CourseListHandler,
    NbGraderVersionHandler,
)
from tornado import web


@contextlib.contextmanager
def chdir(dirname):
    currdir = os.getcwd()
    os.chdir(dirname)
    yield
    os.chdir(currdir)


class E2xAssignmentList(AssignmentList):
    def submit_assignment(self, course_id, assignment_id):
        with self.get_assignment_dir_config() as config:
            try:
                config = self.load_config()
                config.CourseDirectory.course_id = course_id
                config.CourseDirectory.assignment_id = assignment_id

                coursedir = CourseDirectory(config=config)
                authenticator = Authenticator(config=config)
                submit = ExchangeFactory(config=config).Submit(
                    coursedir=coursedir, authenticator=authenticator, config=config
                )

                retval = submit.start()
                hashcode = "Exchange not set up for hashcode"
                timestamp = "Exchange not set up for timestamp"
                if retval and len(retval) == 2:
                    hashcode, timestamp = retval

            except Exception:
                self.log.error(traceback.format_exc())
                retvalue = {"success": False, "value": traceback.format_exc()}

            else:
                retvalue = {
                    "success": True,
                    "hashcode": hashcode,
                    "timestamp": timestamp,
                }

        self.log.info(retvalue)

        return retvalue


class AssignmentActionHandler(BaseAssignmentHandler):
    @web.authenticated
    def post(self, action):
        if action == "fetch":
            assignment_id = self.get_argument("assignment_id")
            course_id = self.get_argument("course_id")
            self.manager.fetch_assignment(course_id, assignment_id)
            self.finish(json.dumps(self.manager.list_assignments(course_id=course_id)))
        elif action == "submit":
            assignment_id = self.get_argument("assignment_id")
            course_id = self.get_argument("course_id")
            output = self.manager.submit_assignment(course_id, assignment_id)
            if output["success"]:
                response = self.manager.list_assignments(course_id=course_id)
                response["hashcode"] = output["hashcode"]
                response["timestamp"] = output["timestamp"]
                self.finish(json.dumps(response))
            else:
                self.finish(json.dumps(output))
        elif action == "fetch_feedback":
            assignment_id = self.get_argument("assignment_id")
            course_id = self.get_argument("course_id")
            self.manager.fetch_feedback(course_id, assignment_id)
            self.finish(json.dumps(self.manager.list_assignments(course_id=course_id)))


# -----------------------------------------------------------------------------
#  URL to handler mappings
# -----------------------------------------------------------------------------


_assignment_action_regex = r"(?P<action>fetch|submit|fetch_feedback)"

default_handlers = [
    (r"/assignments", AssignmentListHandler),
    (r"/assignments/%s" % _assignment_action_regex, AssignmentActionHandler),
    (r"/courses", CourseListHandler),
    (r"/nbgrader_version", NbGraderVersionHandler),
]



from nbgrader.server_extensions.formgrader.handlers import (
    template_path as nbgrader_template_path,
)

from e2xgrader.server_extensions.apps.formgrader.handlers import (
    template_path as e2x_template_path,
)

from ..exporters import E2xExporter

preprocessors = dict(
    AssignLatePenalties="nbgrader.preprocessors.AssignLatePenalties",
    CSSPreprocessor="nbconvert.preprocessors.CSSHTMLHeaderPreprocessor",
    CheckCellMetadata="nbgrader.preprocessors.CheckCellMetadata",
    ClearHiddenTests="e2xgrader.preprocessors.ClearHiddenTests",
    ClearMarkScheme="nbgrader.preprocessors.ClearMarkScheme",
    ClearOutput="nbgrader.preprocessors.ClearOutput",
    ClearSolutions="e2xgrader.preprocessors.ClearSolutions",
    ComputeChecksums="nbgrader.preprocessors.ComputeChecksums",
    DeduplicateIds="nbgrader.preprocessors.DeduplicateIds",
    Execute="nbgrader.preprocessors.Execute",
    ExtractAttachments="e2xgrader.preprocessors.ExtractAttachments",
    FilterTests="e2xgrader.preprocessors.FilterTests",
    GetGrades="nbgrader.preprocessors.GetGrades",
    IncludeHeaderFooter="nbgrader.preprocessors.IncludeHeaderFooter",
    LimitOutput="nbgrader.preprocessors.LimitOutput",
    LockCells="nbgrader.preprocessors.LockCells",
    OverwriteCells="e2xgrader.preprocessors.OverwriteCells",
    OverwriteKernelspec="nbgrader.preprocessors.OverwriteKernelspec",
    SaveAutoGrades="e2xgrader.preprocessors.SaveAutoGrades",
    SaveCells="e2xgrader.preprocessors.SaveCells",
    UnpermuteTasks="e2xgrader.preprocessors.UnpermuteTasks",
    Unscramble="e2xgrader.preprocessors.Unscramble",
    ValidateExtracCells="e2xgrader.preprocessors.ValidateExtraCells",
)


def configure_feedback(config):
    config.GenerateFeedback.preprocessors = [
        preprocessors["GetGrades"],
        preprocessors["FilterTests"],
        preprocessors["CSSPreprocessor"],
    ]
    config.GenerateFeedback.exporter_class = E2xExporter
    config.HTMLExporter.extra_template_basedirs = [
        e2x_template_path,
        nbgrader_template_path,
    ]


def configure_base(config):
    config.Autograde.sanitize_preprocessors = [
        preprocessors["ClearOutput"],
        preprocessors["DeduplicateIds"],
        preprocessors["OverwriteKernelspec"],
        preprocessors["OverwriteCells"],
        preprocessors["CheckCellMetadata"],
        preprocessors["UnpermuteTasks"],
        preprocessors["Unscramble"],
    ]
    config.Autograde.autograde_preprocessors = [
        preprocessors["ExtractAttachments"],
        preprocessors["Execute"],
        preprocessors["LimitOutput"],
        preprocessors["SaveAutoGrades"],
        preprocessors["AssignLatePenalties"],
        preprocessors["CheckCellMetadata"],
    ]
    config.GenerateAssignment.preprocessors = [
        preprocessors["ValidateExtracCells"],
        preprocessors["IncludeHeaderFooter"],
        preprocessors["LockCells"],
        preprocessors["ClearSolutions"],
        preprocessors["ClearOutput"],
        preprocessors["CheckCellMetadata"],
        preprocessors["ComputeChecksums"],
        preprocessors["SaveCells"],
        preprocessors["ClearHiddenTests"],
        preprocessors["ClearMarkScheme"],
        preprocessors["ComputeChecksums"],
        preprocessors["CheckCellMetadata"],
    ]
    configure_feedback(config)


def configure_exchange(config):
    config.ExchangeFactory.exchange = "e2xgrader.exchange.E2xExchange"
    config.ExchangeFactory.submit = "e2xgrader.exchange.E2xExchangeSubmit"
    config.ExchangeFactory.collect = "e2xgrader.exchange.E2xExchangeCollect"
    config.ExchangeFactory.fetch_assignment = (
        "e2xgrader.exchange.E2xExchangeFetchAssignment"
    )
    config.ExchangeFactory.release_assignment = (
        "e2xgrader.exchange.E2xExchangeReleaseAssignment"
    )
    config.ExchangeFactory.list = "e2xgrader.exchange.E2xExchangeList"
    config.ExchangeFactory.fetch_feedback = (
        "e2xgrader.exchange.E2xExchangeFetchFeedback"
    )
    config.ExchangeFactory.release_feedback = (
        "e2xgrader.exchange.E2xExchangeReleaseFeedback"
    )



import hashlib
from textwrap import dedent

import nbformat

from e2xgrader.exporters import E2xExporter


def has_name(cell, name):
    if cell.cell_type != "markdown":
        return False
    return "name" in cell.metadata and cell.metadata.name == name


def append_alert_cell(nb, text, msg, cell_id):
    alert_cell = nbformat.v4.new_markdown_cell()
    alert_cell.source = dedent(
        f"""
        <div class="alert alert-block alert-danger">
            {msg}:
            <h1>{text}</h1>
        </div>
    """
    )
    alert_cell.metadata = {"name": cell_id, "editable": False, "deletable": False}

    # When using notebooks with version <= 4.4 and nbformat v4.5
    # delete the new id attribute to prevent validation errors
    if nb.nbformat == 4 and nb.nbformat_minor <= 4 and "id" in alert_cell:
        del alert_cell["id"]

    target_idx = -1
    for idx, cell in enumerate(nb.cells):
        if has_name(cell, cell_id):
            target_idx = idx
            break

    if target_idx != -1:
        nb.cells[target_idx] = alert_cell
    else:
        nb.cells.append(alert_cell)

    return nb


def append_hashcode(nb, hashcode, msg="Ihr Hashcode"):
    return append_alert_cell(nb, hashcode, msg, "hashcode_cell")


def append_timestamp(nb, timestamp, msg="Timestamp"):
    return append_alert_cell(nb, timestamp, msg, "timestamp_cell")


def compute_hashcode(filename, method="md5"):
    if method == "md5":
        hashcode = hashlib.md5()
    elif method == "sha1":
        hashcode = hashlib.sha1()
    else:
        raise ValueError("Currently only the methods md5 and sha1 are supported!")

    with open(filename, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hashcode.update(chunk)

    return hashcode.hexdigest()


def truncate_hashcode(hashcode, size=20, chunk_size=5):
    hash_string = ""
    for i in range(0, size, chunk_size):
        hash_string += f"-{hashcode[i:i+chunk_size+1]}"
    return hash_string[1:]


def generate_student_info(filename, username, hashcode, timestamp):
    with open(filename, "w") as f:
        f.write(
            dedent(
                f"""
            Username: {username}
            Hashcode: {hashcode}
            Timestamp: {timestamp}
        """
            )
        )


def generate_html(nb, dest):
    exporter = E2xExporter()
    exporter.template_name = "form"
    html, _ = exporter.from_notebook_node(nb)

    with open(dest, "w") as f:
        f.write(html)



from textwrap import dedent

from nbgrader.exchange.default import Exchange
from nbgrader.utils import check_directory
from traitlets import Bool, Unicode


class E2xExchange(Exchange):
    personalized_outbound = Bool(
        False, help="Whether to use a personalized outbound directory per student"
    ).tag(config=True)

    personalized_inbound = Bool(
        False, help="Whether to use a personalized inbound directory per student"
    ).tag(config=True)

    personalized_feedback = Bool(
        False, help="Whether to use a personalized feedback directory per student"
    ).tag(config=True)

    grader = Bool(
        False,
        help=dedent(
            """A flag whether the current user is grader. Used for personalized-outbound
               to retrive the released assignment for one of the students,so that
               lister only shows one released assignment instead of showing all
               individualized assignments from all users
            """
        ),
    ).tag(config=True)

    outbound_directory = Unicode("outbound", help="The name of the outbound directory")

    inbound_directory = Unicode("inbound", help="The name of the inbound directory")

    feedback_directory = Unicode("feedback", help="The name of the feedback directory")

    def __init__(self, coursedir=None, authenticator=None, **kwargs):
        super().__init__(coursedir=coursedir, authenticator=authenticator, **kwargs)

        if self.personalized_outbound:
            self.outbound_directory = "personalized-outbound"

        if self.personalized_inbound:
            self.inbound_directory = "personalized-inbound"

        if self.personalized_feedback:
            self.feedback_directory = "personalized-feedback"

    def ensure_root(self):
        """
        See if the exchange directory exists and readable, fail if not.
        We do not need to make the exchange root writable by default.
        """
        if not check_directory(self.root, read=True, execute=True):
            self.fail(
                "Unreadable directory, please contact your instructor: {}".format(
                    self.root
                )
            )



import base64
import copy
import pickle
import random
import re

import nbformat
from nbgrader.preprocessors import NbGraderPreprocessor


class Scramble(NbGraderPreprocessor):
    def __init__(self, **kw):
        self.__seed = random.randint(0, 10000000)
        self.__random = random.Random(self.__seed)
        self.__p_define = re.compile(
            r"^#define +(?P<fun>\w+)\((?P<args>[^(]*)\) +(?P<body>.*)"
        )
        self.__p_set = re.compile(r"^#set +(?P<name>\w+) += +(?P<options>.*)")
        self.__p_random = re.compile(r"^#random +(?P<vars>.*) +in +(?P<sets>.*)")
        self.__p_replace = re.compile(r"#replace +(?P<name>\w+) +(?P<replace_with>.*)")
        self.__p_lambda = re.compile("^#lambda +(?P<name>[^ ]+) +(?P<lambda>.*)")
        if kw and "seed" in kw:
            self.__random = random.Random(kw["seed"])
            self.__seed = kw["seed"]

    def parse_define(self, line):
        match = self.__p_define.search(line)
        f_name = match.group("fun")
        f_args = [arg.strip() for arg in match.group("args").split(",")]
        f_body = match.group("body")
        return f_name, f_args, f_body

    def parse_set(self, line):
        match = self.__p_set.search(line)
        if match:
            return match.group("name"), match.group("options")
        return None

    def parse_random(self, line, sets):
        match = self.__p_random.search(line)
        if not match:
            return None
        opts = [s.strip() for s in match.group("sets").split(",")]
        var_str = match.group("vars")
        if "!=" in var_str:
            rand_vars = [s.strip() for s in var_str.split("!=")]
            var_groups = [[v.strip() for v in s.split(",")] for s in rand_vars]
            return var_groups, "!=", opts
        else:
            rand_vars = [s.strip() for s in var_str.split("==")]
            var_groups = [[v.strip() for v in s.split(",")] for s in rand_vars]
            return var_groups, "==", opts

    def parse_replace(self, line):
        match = self.__p_replace.search(line)
        if not match:
            return None
        return match.group("name"), match.group("replace_with")

    def parse_lambda(self, line):
        match = self.__p_lambda.search(line)
        if not match:
            return None
        return match.group("name"), "lambda " + match.group("lambda")

    def replace(self, text, macro):
        p_macro = re.compile(r"(?P<fun>{}\((?P<args>[^)]*)\))".format(macro[0]))
        processed = text
        for match in p_macro.finditer(text):
            args = [arg.strip() for arg in match.group("args").split(",")]
            assert len(macro[1]) == len(
                args
            ), "Wrong number of arguments for macro {} with args {}".format(
                macro[0], args
            )
            replacement = macro[2]
            for i in range(len(args)):
                replacement = replacement.replace(macro[1][i], args[i])
            processed = processed.replace(match.group("fun"), replacement)
        return processed

    def replace_lambdas(self, text, name, expr):
        p_macro = re.compile(r"(?P<fun>{}\((?P<args>[^)]*)\))".format(name))
        processed = text
        for match in p_macro.finditer(text):
            args = [arg.strip() for arg in match.group("args").split(",")]

            replacement = expr(*args)
            processed = processed.replace(match.group("fun"), str(replacement))
        return processed

    def sample(self, groups, constraint, sets_names, set_dict):
        sets = [set_dict[sets_name] for sets_name in sets_names]
        rand_dict = {}
        if constraint == "==":
            k = 1
            sampled_idx = self.__random.sample(range(len(sets[0])), k=k)
            for g_idx in range(len(groups)):
                group = groups[g_idx]
                for set_idx in range(len(group)):
                    rand_dict[group[set_idx]] = sets[set_idx][sampled_idx[0]]
            return rand_dict

        if constraint == "!=":
            k = len(groups)
            sampled_idx = self.__random.sample(range(len(sets[0])), k=k)
            for g_idx in range(len(groups)):
                group = groups[g_idx]
                for set_idx in range(len(group)):
                    rand_dict[group[set_idx]] = sets[set_idx][sampled_idx[g_idx]]
            return rand_dict

    def sample_config(self, config):
        lines = config.split("\n")
        new_lines = []
        i = 0
        while i < len(lines):
            line = lines[i].strip()
            while line.endswith("\\"):
                line = line[:-1]
                line += lines[i + 1].strip()
                i += 1
            i += 1

            new_lines.append(line)
        lines = new_lines

        macros = []
        sets = {}
        rands = []
        rand_vars = {}
        replaces = {}
        lambdas = {}
        for line in lines:
            line = line.strip()
            if line.startswith("#define"):
                macros.append(list(self.parse_define(line)))
            elif line.startswith("#set"):
                name, opts = self.parse_set(line)
                sets[name] = [opt.strip() for opt in opts.split("||")]
            elif line.startswith("#random"):
                rands.append(self.parse_random(line, sets))
            elif line.startswith("#replace"):
                name, replace_with = self.parse_replace(line)
                replaces[name] = replace_with
            elif line.startswith("#lambda"):
                name, lambda_expr = self.parse_lambda(line)
                lambdas[name] = eval(lambda_expr)

        for i in range(len(macros)):
            for j in range(i + 1, len(macros)):
                macros[j][2] = self.replace(macros[j][2], macros[i])
            for set_name in sets:
                sets[set_name] = [self.replace(s, macros[i]) for s in sets[set_name]]
            for r_name in replaces:
                replaces[r_name] = self.replace(replaces[r_name], macros[i])

        for rand in rands:
            rand_vars.update(self.sample(rand[0], rand[1], rand[2], sets))

        for rand in rand_vars:
            for r_name in replaces:
                replaces[r_name] = replaces[r_name].replace(rand, rand_vars[rand])

        for rand in rand_vars:
            for r_name in replaces:
                replaces[r_name] = replaces[r_name].replace(rand, rand_vars[rand])

        for rand in rand_vars:
            for r_name in replaces:
                replaces[r_name] = replaces[r_name].replace(rand, rand_vars[rand])

        for lambda_expr in lambdas:
            for r_name in replaces:
                replaces[r_name] = self.replace_lambdas(
                    replaces[r_name], lambda_expr, lambdas[lambda_expr]
                )

        return {
            "seed": self.__seed,
            "macros": macros,
            "sets": sets,
            "rands": rand_vars,
            "replace": replaces,
        }

    def obscure(self, my_dict):
        byte_str = pickle.dumps(my_dict)
        obscured = base64.b85encode(byte_str)
        return obscured

    def preprocess(self, nb, resources):
        if len(nb.cells) < 1 or not nb.cells[0].source.startswith("%% scramble"):
            return nb, resources
        config = self.sample_config(nb.cells[0].source)
        replacement_variables = config["replace"]
        scrambled_nb = nbformat.v4.new_notebook()
        scrambled_nb.cells = copy.deepcopy(nb.cells[1:])
        for cell in scrambled_nb.cells:
            for replacement_variable in replacement_variables:
                cell.source = cell.source.replace(
                    "{{" + replacement_variable + "}}",
                    replacement_variables[replacement_variable],
                )
        scrambled_nb.metadata["scramble_config"] = {
            "seed": config["seed"],
            "config": self.obscure(replacement_variables),
        }
        return scrambled_nb, resources



import json

from e2xcore.handlers.base import E2xApiHandler
from nbgrader.server_extensions.formgrader.apihandlers import (
    AssignmentCollectionHandler,
    AssignmentHandler,
)
from nbgrader.server_extensions.formgrader.base import check_notebook_dir, check_xsrf
from tornado import web

from e2xgrader.exporters import GradeExporter


class E2xAssignmentCollectionHandler(E2xApiHandler, AssignmentCollectionHandler):
    """
    Inherit from E2xApiHandler to overwrite the internal NbgraderAPI with the E2xAPI
    """

    @web.authenticated
    @check_xsrf
    @check_notebook_dir
    def get(self):
        include_score = self.get_argument("include_score", "True").lower() == "true"
        self.write(json.dumps(self.api.get_assignments(include_score=include_score)))


class E2xAssignmentHandler(E2xApiHandler, AssignmentHandler):
    """
    Inherit from E2xApiHandler to overwrite the internal NbgraderAPI with the E2xAPI
    """

    @web.authenticated
    @check_xsrf
    @check_notebook_dir
    def get(self, assignment_id):
        include_score = self.get_argument("include_score", "True").lower() == "true"
        assignment = self.api.get_assignment(assignment_id, include_score=include_score)
        if assignment is None:
            raise web.HTTPError(404)
        self.write(json.dumps(assignment))


class SolutionCellCollectionHandler(E2xApiHandler):
    @web.authenticated
    @check_xsrf
    def get(self, assignment_id, notebook_id):
        cells = self.api.get_solution_cell_ids(assignment_id, notebook_id)
        self.write(json.dumps(cells))


class SubmittedTaskCollectionHandler(E2xApiHandler):
    @web.authenticated
    @check_xsrf
    def get(self, assignment_id, notebook_id, task_id):
        submissions = self.api.get_task_submissions(assignment_id, notebook_id, task_id)
        self.write(json.dumps(submissions))


class GenerateFeedbackHandler(E2xApiHandler):
    @web.authenticated
    @check_xsrf
    def post(self, assignment_id, student_id):
        hide_cells = json.loads(self.get_argument("hide_cells", "false"))
        self.write(
            json.dumps(
                self.api.generate_feedback(
                    assignment_id, student_id, hide_cells=hide_cells
                )
            )
        )


class GenerateAllFeedbackHandler(E2xApiHandler):
    @web.authenticated
    @check_xsrf
    def post(self, assignment_id):
        hide_cells = json.loads(self.get_argument("hide_cells", "false"))
        self.write(
            json.dumps(self.api.generate_feedback(assignment_id, hide_cells=hide_cells))
        )


class GraderHandler(E2xApiHandler):
    @web.authenticated
    @check_xsrf
    def get(self):
        e2xgrader_settings = self.settings.get("e2xgrader", dict())
        grader_settings = e2xgrader_settings.get("graders", list())
        self.write(json.dumps(grader_settings))


class ExportGradesHandler(E2xApiHandler):
    def initialize(self) -> None:
        self.__exporter = GradeExporter()
        self.set_header("Content-Type", 'text/csv; charset="utf-8"')
        self.set_header("Content-Disposition", "attachment; filename=grades.csv")

    @web.authenticated
    @check_xsrf
    def get(self):
        level = self.get_argument("level", "notebook")
        if level == "task":
            self.__exporter.tasks = True
        elif level == "assignment":
            self.__exporter.notebooks = False
        normalize = self.get_argument("normalize", "false")
        if normalize.lower() == "max_score":
            self.__exporter.include_max_score = True
        else:
            self.__exporter.normalize = normalize.lower() == "true"
        self.write(self.__exporter.get_grades().to_csv())
        self.finish()


class AnnotationCollectionHandler(E2xApiHandler):
    @web.authenticated
    @check_xsrf
    @check_notebook_dir
    def get(self):
        solution_cells = self.api.get_annotations(self.get_argument("submission_id"))
        if solution_cells is None:
            raise web.HTTPError(404)
        return self.write(json.dumps(solution_cells))


class AnnotationHandler(E2xApiHandler):
    @web.authenticated
    @check_xsrf
    @check_notebook_dir
    def put(self, solution_cell_id):
        data = self.get_json_body()
        solution_cell = self.api.save_annotation(
            submission_id=data.get("submission_id"),
            name=data.get("name"),
            annotation=data.get("annotation"),
        )
        if solution_cell is None:
            raise web.HTTPError(404)

        self.write(json.dumps(solution_cell))


default_handlers = [
    (r"/formgrader/api/assignments", E2xAssignmentCollectionHandler),
    (r"/formgrader/api/assignment/([^/]+)", E2xAssignmentHandler),
    (r"/formgrader/api/solution_cells/([^/]+)/([^/]+)", SolutionCellCollectionHandler),
    (
        r"/formgrader/api/submitted_tasks/([^/]+)/([^/]+)/([^/]+)",
        SubmittedTaskCollectionHandler,
    ),
    (
        r"/formgrader/api/assignment/([^/]+)/generate_feedback",
        GenerateAllFeedbackHandler,
    ),
    (
        r"/formgrader/api/assignment/([^/]+)/([^/]+)/generate_feedback",
        GenerateFeedbackHandler,
    ),
    (
        r"/formgrader/api/export_grades/?",
        ExportGradesHandler,
    ),
    (r"/formgrader/api/annotations", AnnotationCollectionHandler),
    (r"/formgrader/api/annotation/([^/]+)", AnnotationHandler),
]



import glob
import os
from textwrap import dedent

import nbformat
from nbgrader.exchange.default import ExchangeFetchAssignment
from nbgrader.utils import check_mode

from ..preprocessors import PermuteTasks, Scramble
from .exchange import E2xExchange


class E2xExchangeFetchAssignment(E2xExchange, ExchangeFetchAssignment):
    def init_src(self):
        if self.coursedir.course_id == "":
            self.fail("No course id specified. Re-run with --course flag.")
        if not self.authenticator.has_access(
            self.coursedir.student_id, self.coursedir.course_id
        ):
            self.fail("You do not have access to this course.")

        self.course_path = os.path.join(self.root, self.coursedir.course_id)
        self.outbound_path = os.path.join(self.course_path, self.outbound_directory)
        if self.personalized_outbound:
            self.src_path = os.path.join(
                self.outbound_path,
                os.getenv("JUPYTERHUB_USER"),
                self.coursedir.assignment_id,
            )
        else:
            self.src_path = os.path.join(
                self.outbound_path, self.coursedir.assignment_id
            )

        if not os.path.isdir(self.src_path):
            self._assignment_not_found(
                self.src_path, os.path.join(self.outbound_path, "*")
            )
        if not check_mode(self.src_path, read=True, execute=True):
            self.fail(
                "You don't have read permissions for the directory: {}".format(
                    self.src_path
                )
            )

    def do_scrambling(self, dest, student_id):
        self.log.info(f"Scrambling for {student_id}")
        scrambler = Scramble(seed=hash(student_id))
        permuter = PermuteTasks(seed=hash(student_id))
        for nb_path in glob.glob(os.path.join(dest, "*.ipynb")):
            nb = nbformat.read(nb_path, as_version=4)
            if len(nb.cells) > 0 and nb.cells[0].source.startswith("%% scramble"):
                resources = {}
                nb, resources = scrambler.preprocess(nb, resources)
                nb, resources = permuter.preprocess(nb, resources)
                nbformat.write(nb, nb_path)
        self.log.info("Scrambled")

    def copy_files(self):
        self.log.info("Source: {}".format(self.src_path))
        self.log.info("Destination: {}".format(self.dest_path))
        self.do_copy(self.src_path, self.dest_path)
        self.do_scrambling(self.dest_path, os.getenv("JUPYTERHUB_USER"))
        self.log.info(
            "Fetched as: {} {}".format(
                self.coursedir.course_id, self.coursedir.assignment_id
            )
        )

    def do_copy(self, src, dest):
        if self.personalized_outbound:
            personalized_src = os.path.join(src, os.getenv("JUPYTERHUB_USER"))
            if os.path.exists(personalized_src):
                src = personalized_src
            else:
                self.log.warning(
                    dedent(
                        f"""
                    Using personalized outbound, but no directory for
                    user {os.getenv('JUPYTERHUB_USER')} exists.
                    """
                    )
                )
        super().do_copy(src, dest)



from traitlets import Any, List

from ..apps.assignment_list import AssignmentList
from ..apps.help import Help
from ..apps.validate_assignment import ValidateAssignment
from ..base import BaseExtension


class StudentExtension(BaseExtension):
    apps = List(
        trait=Any(),
        default_value=[AssignmentList, ValidateAssignment, Help],
    ).tag(config=True)


def load_jupyter_server_extension(nbapp):
    """Load the e2xgrader serverextension"""
    nbapp.log.info("Loading the e2xgrader student serverextension")
    StudentExtension(parent=nbapp)



import glob
import os

from nbgrader.exchange.default import ExchangeFetchFeedback
from nbgrader.utils import check_mode, get_username, make_unique_key, notebook_hash

from .exchange import E2xExchange


class E2xExchangeFetchFeedback(E2xExchange, ExchangeFetchFeedback):
    def init_src(self):
        if self.coursedir.course_id == "":
            self.fail("No course id specified. Re-run with --course flag.")

        self.course_path = os.path.join(self.root, self.coursedir.course_id)
        self.outbound_path = os.path.join(self.course_path, self.feedback_directory)
        self.src_path = os.path.join(self.outbound_path)
        self.cache_path = os.path.join(self.cache, self.coursedir.course_id)

        if self.coursedir.student_id != "*":
            # An explicit student id has been specified on the command line; we use it as student_id
            if "*" in self.coursedir.student_id or "+" in self.coursedir.student_id:
                self.fail(
                    "The student ID should contain no '*' nor '+'; got {}".format(
                        self.coursedir.student_id
                    )
                )
            student_id = self.coursedir.student_id
        else:
            student_id = get_username()

        if not os.path.isdir(self.src_path):
            self._assignment_not_found(
                self.src_path, os.path.join(self.outbound_path, "*")
            )
        if not check_mode(self.src_path, execute=True):
            self.fail(
                "You don't have execute permissions for the directory: {}".format(
                    self.src_path
                )
            )

        assignment_id = (
            self.coursedir.assignment_id if self.coursedir.assignment_id else "*"
        )
        pattern = os.path.join(self.cache_path, "*+{}+*".format(assignment_id))
        self.log.debug("Looking for submissions with pattern: {}".format(pattern))

        self.feedback_files = []
        submissions = [os.path.split(x)[-1] for x in glob.glob(pattern)]
        for submission in submissions:
            _, assignment_id, timestamp = submission.split("/")[-1].split("+")

            self.log.debug(
                "Looking for feedback for '{}/{}' submitted at {}".format(
                    self.coursedir.course_id, assignment_id, timestamp
                )
            )

            pattern = os.path.join(self.cache_path, submission, "*.ipynb")
            notebooks = glob.glob(pattern)
            for notebook in notebooks:
                notebook_id = os.path.splitext(os.path.split(notebook)[-1])[0]

                # Check if personalized_feedback is used
                if self.personalized_feedback:
                    feedbackpath = os.path.join(
                        self.outbound_path,
                        student_id,
                        assignment_id,
                        "{}.html".format(notebook_id),
                    )
                    self.log.debug("Feedback file: ", feedbackpath)
                    if os.path.exists(feedbackpath):
                        self.feedback_files.append(
                            (notebook_id, timestamp, feedbackpath)
                        )
                        self.log.info(
                            "Found feedback for '{}/{}/{}' submitted at {}".format(
                                self.coursedir.course_id,
                                assignment_id,
                                notebook_id,
                                timestamp,
                            )
                        )
                        continue

                    # If we reached here, then there's no feedback available
                    self.log.warning(
                        "Could not find feedback for '{}/{}/{}' submitted at {}".format(
                            self.coursedir.course_id,
                            assignment_id,
                            notebook_id,
                            timestamp,
                        )
                    )
                else:
                    unique_key = make_unique_key(
                        self.coursedir.course_id,
                        assignment_id,
                        notebook_id,
                        student_id,
                        timestamp,
                    )

                    # Look for the feedback using new-style of feedback
                    self.log.debug("Unique key is: {}".format(unique_key))
                    nb_hash = notebook_hash(notebook, unique_key)
                    feedbackpath = os.path.join(
                        self.outbound_path, "{0}.html".format(nb_hash)
                    )
                    if os.path.exists(feedbackpath):
                        self.feedback_files.append(
                            (notebook_id, timestamp, feedbackpath)
                        )
                        self.log.info(
                            "Found feedback for '{}/{}/{}' submitted at {}".format(
                                self.coursedir.course_id,
                                assignment_id,
                                notebook_id,
                                timestamp,
                            )
                        )
                        continue

                    # If it doesn't exist, try the legacy hashing
                    nb_hash = notebook_hash(notebook)
                    feedbackpath = os.path.join(
                        self.outbound_path, "{0}.html".format(nb_hash)
                    )
                    if os.path.exists(feedbackpath):
                        self.feedback_files.append(
                            (notebook_id, timestamp, feedbackpath)
                        )
                        self.log.warning(
                            "Found legacy feedback for '{}/{}/{}' submitted at {}".format(
                                self.coursedir.course_id,
                                assignment_id,
                                notebook_id,
                                timestamp,
                            )
                        )
                        continue

                    # If we reached here, then there's no feedback available
                    self.log.warning(
                        "Could not find feedback for '{}/{}/{}' submitted at {}".format(
                            self.coursedir.course_id,
                            assignment_id,
                            notebook_id,
                            timestamp,
                        )
                    )



from e2xcore import BaseApp

from .handlers import default_handlers


class ValidateAssignment(BaseApp):
    def load_app(self):
        self.log.info("Loading the e2x validate assignment app")
        self.add_handlers(default_handlers)



from logging import Logger
from typing import Optional, Tuple

from nbformat.notebooknode import NotebookNode
from nbgrader.utils import compute_checksum, is_solution
from traitlets.config import LoggingConfigurable


class BaseGrader(LoggingConfigurable):
    def cell_changed(self, cell: NotebookNode):
        return not (
            is_solution(cell)
            and "checksum" in cell.metadata.nbgrader
            and cell.metadata.nbgrader["checksum"] == compute_checksum(cell)
        )

    def determine_grade(
        self, cell: NotebookNode, log: Logger = None
    ) -> Tuple[Optional[float], float]:
        max_points = float(cell.metadata["nbgrader"]["points"])
        return None, max_points



from typing import Tuple

from nbconvert.exporters.exporter import ResourcesDict
from nbformat.notebooknode import NotebookNode
from nbgrader.preprocessors import ClearHiddenTests as NbgraderClearHiddenTests

from ..utils.extra_cells import clear_choices, is_multiplechoice, is_singlechoice


class ClearHiddenTests(NbgraderClearHiddenTests):
    def preprocess_cell(
        self, cell: NotebookNode, resources: ResourcesDict, cell_index: int
    ) -> Tuple[NotebookNode, ResourcesDict]:
        if is_singlechoice(cell) or is_multiplechoice(cell):
            clear_choices(cell)
            return cell, resources
        else:
            return super().preprocess_cell(cell, resources, cell_index)



import base64
import pickle
import unittest
from copy import deepcopy

from nbformat.v4 import new_notebook

from e2xgrader.preprocessors import Unscramble
from e2xgrader.tests.test_utils.cells import (
    new_manually_graded_code_cell,
    new_manually_graded_markdown_cell,
    new_readonly_code_cell,
)


class TestUnscramble(unittest.TestCase):
    def setUp(self) -> None:
        self.config = dict(myvar1="5", myvar2="Hello", somevar="Goodbye!")

        self.nb = new_notebook(
            metadata=dict(
                scramble_config=dict(
                    seed=1289341782, config=base64.b85encode(pickle.dumps(self.config))
                )
            )
        )

        self.nb.cells.extend(
            [
                new_manually_graded_markdown_cell(
                    source="Here we want to replace {{ myvar1   }}!"
                ),
                new_readonly_code_cell(source="# Here we want to replace {{myvar2 }}"),
                new_manually_graded_code_cell(
                    source=(
                        "# Here we want to replace everything {{myvar1}} "
                        "{{ myvar2 }} and {{ somevar }}"
                    )
                ),
            ]
        )

    def test_unscramble(self):
        nb = deepcopy(self.nb)
        processed_nb, _ = Unscramble().preprocess(nb, {})
        for cell in processed_nb.cells:
            # Make sure none of the original names are there anymore
            assert all([key not in cell.source for key in self.config.keys()])
        for original_cell, processed_cell in zip(self.nb.cells, processed_nb.cells):
            for key, value in self.config.items():
                if key in original_cell:
                    assert key not in processed_cell
                    assert value in processed_cell

    def test_no_scramble_config(self):
        nb = deepcopy(self.nb)
        del nb.metadata["scramble_config"]
        processed_nb, _ = Unscramble().preprocess(nb, {})
        for original_cell, processed_cell in zip(self.nb.cells, processed_nb.cells):
            assert original_cell.source == processed_cell.source



from .base import BaseExtensionManager
from .utils import discover_nbextensions, get_nbextension_utils


class NbExtensionManager(BaseExtensionManager):
    def __init__(self):
        super().__init__()
        self.utils = get_nbextension_utils()
        if self.utils is None:
            self.log.warn(
                "Neither notebook<7 or nbclassic found. Won't install nbextensions."
            )
        self.modules = ["e2xgrader", "nbgrader"]

    def enable_nbextensions(
        self, mode: str, sys_prefix: bool = True, user: bool = False
    ) -> None:
        for nbextension in discover_nbextensions(mode):
            self.utils.enable_nbextension(
                **nbextension, sys_prefix=sys_prefix, user=user
            )

    def deactivate(self, sys_prefix: bool = True, user: bool = False) -> None:
        if self.utils is None:
            return
        for module in self.modules:
            kwargs = dict(module=module, sys_prefix=sys_prefix, user=user)
            self.utils.install_nbextension_python(**kwargs)
            self.utils.disable_nbextension_python(**kwargs)
            self.utils.uninstall_nbextension_python(**kwargs)

    def activate_common(self, sys_prefix: bool = True, user: bool = False) -> None:
        self.deactivate(sys_prefix=sys_prefix, user=user)
        for module in self.modules:
            kwargs = dict(module=module, sys_prefix=sys_prefix, user=user)
            self.utils.install_nbextension_python(**kwargs)
            self.utils.disable_nbextension_python(**kwargs)

    def activate_teacher(self, sys_prefix: bool = True, user: bool = False) -> None:
        if self.utils is None:
            return
        self.activate_common(sys_prefix=sys_prefix, user=user)
        self.enable_nbextensions(mode="teacher", sys_prefix=sys_prefix, user=user)
        self.utils.enable_nbextension_python(
            "nbgrader", sys_prefix=sys_prefix, user=user
        )
        self.utils.disable_nbextension(
            require="create_assignment/main",
            section="notebook",
            sys_prefix=sys_prefix,
            user=user,
        )

    def activate_student(self, sys_prefix: bool = True, user: bool = False) -> None:
        if self.utils is None:
            return
        self.activate_common(sys_prefix=sys_prefix, user=user)
        self.enable_nbextensions(mode="student", sys_prefix=sys_prefix, user=user)
        self.utils.enable_nbextension(
            require="assignment_list/main",
            section="tree",
            sys_prefix=sys_prefix,
            user=user,
        )

    def activate_student_exam(
        self, sys_prefix: bool = True, user: bool = False
    ) -> None:
        if self.utils is None:
            return
        self.activate_common(sys_prefix=sys_prefix, user=user)
        self.enable_nbextensions(mode="student_exam", sys_prefix=sys_prefix, user=user)
        self.utils.enable_nbextension(
            require="assignment_list/main",
            section="tree",
            sys_prefix=sys_prefix,
            user=user,
        )



"""
A system for creating assignments.
"""

import glob
import os
from os.path import join as pjoin

from .server_extensions.student import (
    _load_jupyter_server_extension as load_student_extension,
)
from .server_extensions.student_exam import (
    _load_jupyter_server_extension as load_student_exam_extension,
)
from .server_extensions.teacher import (
    _load_jupyter_server_extension as load_teacher_extension,
)


def _jupyter_nbextension_paths():
    root = os.path.dirname(__file__)
    base_path = pjoin(root, "static", "nbextensions")

    paths = []

    for section in ["notebook", "tree"]:
        for notebook_extension in glob.glob(pjoin(base_path, section, "*")):
            name = f"{os.path.split(notebook_extension)[-1]}_{section}"
            paths.append(
                dict(
                    section=section,
                    src=notebook_extension,
                    dest=name,
                    require=pjoin(name, "main"),
                )
            )

    return paths


def _jupyter_server_extension_paths():
    paths = [
        dict(module="e2xgrader.server_extensions.teacher"),
        dict(module="e2xgrader.server_extensions.student"),
        dict(module="e2xgrader.server_extensions.student_exam"),
    ]

    return paths


def _jupyter_server_extension_points():
    return [dict(module="e2xgrader")]


def _load_jupyter_server_extension(app):
    load_student_extension(app)
    load_student_exam_extension(app)
    load_teacher_extension(app)



import unittest

from bs4 import BeautifulSoup

from e2xgrader.exporters.filters import Highlight2HTMLwithLineNumbers


class TestHighlight2HTMLwithLineNumbers(unittest.TestCase):
    def test_insert_line_numbers(self):
        source = "line 1\nline 2\nline 3"
        highlighter = Highlight2HTMLwithLineNumbers()
        html = highlighter.__call__(source)

        soup = BeautifulSoup(html, features="html.parser")
        line_numbers = soup.find_all("span", attrs={"class": "linenos"})
        assert len(line_numbers) == len(source.split("\n"))

    def test_insert_line_numbers_with_language(self):
        source = "line 1\nline 2\nline 3"
        highlighter = Highlight2HTMLwithLineNumbers()
        html = highlighter(source, language="ipython3")

        soup = BeautifulSoup(html, features="html.parser")
        line_numbers = soup.find_all("span", attrs={"class": "linenos"})
        assert len(line_numbers) == len(source.split("\n"))



from .base import BaseGrader
from .code import CodeGrader
from .multiplechoice import MultipleChoiceGrader
from .singlechoice import SingleChoiceGrader

__all__ = ["BaseGrader", "CodeGrader", "MultipleChoiceGrader", "SingleChoiceGrader"]



from nbformat.v4 import new_code_cell, new_markdown_cell
from nbgrader.utils import is_grade, is_locked, is_solution


def new_read_only_cell(grade_id, source=None, cell_type="markdown"):
    if source is None:
        source = ""
    if cell_type == "markdown":
        cell = new_markdown_cell(source)
    elif cell_type == "code":
        cell = new_code_cell(source)
    else:
        raise NameError(f"{cell_type} is not a supported cell type!")
    cell.metadata["nbgrader"] = {
        "grade": False,
        "grade_id": grade_id,
        "locked": True,
        "schema_version": 3,
        "solution": False,
        "task": False,
    }
    return cell


def is_nbgrader_cell(cell):
    return "nbgrader" in cell.metadata


def is_solution_cell(cell):
    return is_nbgrader_cell(cell) and is_solution(cell)


def is_description(cell):
    return is_nbgrader_cell(cell) and not is_grade(cell) and is_locked(cell)


def grade_id(cell):
    if is_nbgrader_cell(cell):
        return cell.metadata.nbgrader.grade_id


def get_points(cell):
    if is_grade(cell):
        return cell.metadata.nbgrader.points
    return 0


def get_valid_name(name):
    chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
    others = "01234567890_-"
    invalid = ""
    # Make sure at least one character is present
    if not any(char in chars for char in name):
        name = "Task_{}".format(name)
    # Identify and replace invalid chars
    for char in name:
        if char not in chars + others:
            invalid += char
    for char in invalid:
        name = name.replace(char, "_")
    return name


def get_task_info(nb):
    subtasks = []
    subtask = []
    for idx, cell in enumerate(nb.cells):
        subtask.append(idx)
        if is_grade(cell):
            subtasks.append(subtask)
            subtask = []
    task = dict()
    if (
        len(subtasks) > 0
        and len(subtasks[0]) > 0
        and is_description(nb.cells[subtasks[0][0]])
    ):
        task["header"] = subtasks[0].pop(0)
    task["subtasks"] = subtasks
    if len(subtask) > 0:
        task["other"] = subtask
    return task


def get_tasks(nb):
    task_ids = [grade_id(cell) for cell in nb.cells if is_solution(cell)]
    associated = dict()
    checked = dict()

    for task_id in task_ids:
        checked[task_id] = False
        associated[task_id] = []
        for cell in nb.cells:
            if is_nbgrader_cell(cell) and task_id in grade_id(cell):
                associated[task_id].append(grade_id(cell))

    groups = []
    for i in range(len(task_ids)):
        task_id = task_ids[i]
        if checked[task_id]:
            continue
        new_group = set(associated[task_id])
        for j in range(i + 1, len(task_ids)):
            task_idj = task_ids[j]
            if any(t in associated[task_idj] for t in associated[task_id]):
                new_group.update(associated[task_idj])
                checked[task_idj] = True
        groups.append(new_group)

    cell_groups = []
    for group in groups:
        cell_group = []
        for i in range(len(nb.cells)):
            cell = nb.cells[i]
            if is_nbgrader_cell(cell) and grade_id(cell) in group:
                cell_group.append(i)
        cell_groups.append(cell_group)

    return cell_groups



from logging import Logger
from typing import Optional, Tuple

from nbformat.notebooknode import NotebookNode

from ..utils.extra_cells import get_choices, get_instructor_choices
from .base import BaseGrader


class SingleChoiceGrader(BaseGrader):
    def determine_grade(
        self, cell: NotebookNode, log: Logger = None
    ) -> Tuple[Optional[float], float]:
        max_points = float(cell.metadata["nbgrader"]["points"])
        student_choices = get_choices(cell)
        instructor_choices = get_instructor_choices(cell)

        if (
            (len(student_choices) > 0)
            and (len(instructor_choices) > 0)
            and (student_choices[0] == instructor_choices[0])
        ):
            return max_points, max_points

        return 0, max_points



from .student_exam import load_jupyter_server_extension

_load_jupyter_server_extension = load_jupyter_server_extension

__all__ = [
    "_load_jupyter_server_extension",
    "load_jupyter_server_extension",
]



import unittest

from e2xgrader.preprocessors import ClearSolutions
from e2xgrader.tests.test_utils.cells import (
    new_diagram_cell,
    new_manually_graded_markdown_cell,
    new_singlechoice_cell,
)


class TestClearSolutions(unittest.TestCase):
    def setUp(self) -> None:
        self.resources = dict(language="python")

    def test_markdown_cell(self):
        processed_cell, _ = ClearSolutions().preprocess_cell(
            cell=new_manually_graded_markdown_cell(source="This is the answer!"),
            resources=self.resources,
            cell_index=0,
        )
        assert processed_cell.source == "YOUR ANSWER HERE"

    def test_extra_cell(self):
        source = "My original text"
        cell = new_singlechoice_cell(source=source)

        processed_cell, _ = ClearSolutions().preprocess_cell(
            cell=cell,
            resources=self.resources,
            cell_index=0,
        )
        assert processed_cell.source == source

    def test_diagram_cell_no_remove(self):
        cell = new_diagram_cell(
            options=dict(replace_diagram=dict(value=False)),
            diagram="some_base64_encoded_data",
        )
        processed_cell, _ = ClearSolutions().preprocess_cell(
            cell=cell, resources=self.resources, cell_index=0
        )
        assert "diagram.png" in processed_cell.attachments

    def test_diagram_cell_remove(self):
        cell = new_diagram_cell(
            options=dict(replace_diagram=dict(value=True)),
            diagram="some_base64_encoded_data",
        )
        processed_cell, _ = ClearSolutions().preprocess_cell(
            cell=cell, resources=self.resources, cell_index=0
        )
        assert "diagram.png" not in processed_cell.attachments



import unittest

from e2xgrader.extensions.utils import is_installed


class TestUtils(unittest.TestCase):
    def test_is_installed(self):
        self.assertTrue(is_installed("e2xgrader"), "e2xgrader should be installed")
        self.assertFalse(
            is_installed("e2xgradeeer"),
            "Non existing package e2xgradeeer should not be installed",
        )



import base64
import glob
import os
import os.path

from bs4 import BeautifulSoup
from jinja2.filters import pass_context
from nbconvert.exporters import HTMLExporter
from nbgrader.server_extensions.formgrader import handlers as nbgrader_handlers
from traitlets import Unicode

from ..utils import extra_cells as utils
from .filters import Highlight2HTMLwithLineNumbers


class E2xExporter(HTMLExporter):
    """
    Custom E2x notebook exporter
    """

    extra_cell_field = Unicode(
        "extended_cell", help="The name of the extra cell metadata field."
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.extra_template_basedirs = [
            os.path.abspath(
                os.path.join(
                    os.path.dirname(__file__),
                    "..",
                    "server_extensions",
                    "apps",
                    "e2xgraderapi",
                    "templates",
                )
            ),
            os.path.abspath(
                os.path.join(
                    os.path.dirname(__file__),
                    "..",
                    "server_extensions",
                    "apps",
                    "formgrader",
                    "templates",
                )
            ),
            nbgrader_handlers.template_path,
        ]
        # The notebook seems to sometimes set exclude_input to true
        self.exclude_input = False

    @property
    def template_paths(self):
        return super()._template_paths() + [
            os.path.abspath(
                os.path.join(
                    os.path.dirname(__file__),
                    "..",
                    "server_extensions",
                    "grader",
                    "apps",
                    "formgrader",
                    "templates",
                )
            )
        ]

    @pass_context
    def to_choicecell(self, context, source):
        cell = context.get("cell", {})
        soup = BeautifulSoup(source, "html.parser")
        my_type = None
        if not soup.ul or not utils.is_extra_cell(cell):
            return soup.prettify().replace("\n", "")
        if utils.is_singlechoice(cell):
            my_type = "radio"
        elif utils.is_multiplechoice(cell):
            my_type = "checkbox"
        form = soup.new_tag("form")
        form["class"] = "hbrs_checkbox"

        list_elems = soup.ul.find_all("li")
        for i in range(len(list_elems)):
            div = soup.new_tag("div")
            box = soup.new_tag("input")
            box["type"] = my_type
            box["value"] = i
            box["disabled"] = "disabled"
            if i in utils.get_choices(cell):
                box["checked"] = "checked"
            div.append(box)
            children = [c for c in list_elems[i].children]
            for child in children:
                div.append(child)

            if utils.has_solution(cell):
                check = soup.new_tag("span")
                if i in utils.get_instructor_choices(cell):
                    check.string = "correct"
                    check["style"] = "color:green"
                else:
                    check.string = "false"
                    check["style"] = "color:red"
                div.append(check)
            form.append(div)
        soup.ul.replaceWith(form)
        return soup.prettify().replace("\n", "")

    def default_filters(self):
        for pair in super(E2xExporter, self).default_filters():
            yield pair
        yield ("to_choicecell", self.to_choicecell)

    def discover_annotations(self, resources):
        if resources is None:
            return
        resources["annotations"] = dict()
        if "metadata" not in resources or "path" not in resources["metadata"]:
            return

        path = resources["metadata"]["path"]

        for annotation in glob.glob(os.path.join(path, "annotations", "*.png")):
            cell_id = os.path.splitext(os.path.basename(annotation))[0]
            with open(annotation, "rb") as f:
                img = base64.b64encode(f.read()).decode("utf-8")
                resources["annotations"][cell_id] = f"data:image/png;base64,{img}"

    def from_notebook_node(self, nb, resources=None, **kw):
        self.discover_annotations(resources)

        self.exclude_input = False
        langinfo = nb.metadata.get("language_info", {})
        lexer = langinfo.get("pygments_lexer", langinfo.get("name", None))
        highlight_code = self.filters.get(
            "highlight_code_with_linenumbers",
            Highlight2HTMLwithLineNumbers(pygments_lexer=lexer, parent=self),
        )
        self.register_filter("highlight_code_with_linenumbers", highlight_code)
        return super(E2xExporter, self).from_notebook_node(nb, resources, **kw)



from importlib import import_module

from jinja2 import Environment, FileSystemLoader
from traitlets import Any, List, TraitError, validate
from traitlets.config import Application

from e2xgrader.utils import get_nbgrader_config


class BaseExtension(Application):
    apps = List(
        trait=Any(),
        default_value=[],
    ).tag(config=True)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.log = self.parent.log
        self.config = get_nbgrader_config()
        self.log.info(self.apps)
        self.initialize_jinja_environment()
        self.initialize_apps()

    @validate("apps")
    def validate_apps(self, proposal):
        apps = []
        for app in proposal["value"]:
            if isinstance(app, str):
                module, klass = app.rsplit(".", 1)
                app = getattr(import_module(module), klass)
            if not callable(app):
                raise TraitError("apps must be callable")
            apps.append(app)
        return apps

    def initialize_apps(self):
        for app in self.apps:
            app(parent=self.parent).load_app()

    def initialize_jinja_environment(self):
        self.parent.web_app.settings["e2xgrader"] = {
            "jinja_env": Environment(loader=FileSystemLoader([]))
        }



import unittest

from nbformat.notebooknode import NotebookNode
from nbformat.v4 import new_code_cell
from nbgrader.utils import compute_checksum

from e2xgrader.graders import BaseGrader


class TestBaseGrader(unittest.TestCase):
    def create_unchanged_code_cell(self, grade_id: str) -> NotebookNode:
        cell = new_code_cell()

        cell.metadata = {
            "nbgrader": {
                "grade": True,
                "grade_id": grade_id,
                "locked": False,
                "points": 5,
                "schema_version": 3,
                "solution": True,
                "task": False,
            }
        }

        cell.source = "YOUR CODE HERE"

        cell.metadata.nbgrader["checksum"] = compute_checksum(cell)

        return cell

    def create_changed_code_cell(self, grade_id: str) -> NotebookNode:
        cell = new_code_cell()

        cell.metadata = {
            "nbgrader": {
                "grade": True,
                "grade_id": grade_id,
                "locked": False,
                "points": 5,
                "schema_version": 3,
                "solution": True,
                "task": False,
            }
        }

        cell.source = "YOUR CODE HERE"

        cell.metadata.nbgrader["checksum"] = compute_checksum(cell)

        cell.source = "MY SOLUTION"

        return cell

    def test_cell_unchanged(self):
        assert not BaseGrader().cell_changed(
            self.create_unchanged_code_cell("mycell")
        ), "cell should not have changed"

    def test_cell_changed(self):
        assert BaseGrader().cell_changed(
            self.create_changed_code_cell("mycell")
        ), "cell should have changed"



from typing import List, Union

from nbformat.notebooknode import NotebookNode
from nbgrader import utils
from nbgrader.validator import Validator

from ....utils.extra_cells import (
    get_choices,
    is_attachment_cell,
    is_extra_cell,
    is_multiplechoice,
    is_singlechoice,
)


class E2XValidator(Validator):
    def validate_extra_cell(self, cell: NotebookNode) -> Union[None, float]:
        if (
            (is_multiplechoice(cell) or is_singlechoice(cell))
            and len(get_choices(cell)) < 1
        ) or (is_attachment_cell(cell) and len(cell.get("attachments", {})) < 1):
            return 0

    def code_cell_errored(self, cell: NotebookNode) -> bool:
        for output in cell.outputs:
            if (
                output.output_type == "error"
                or output.output_type == "stream"
                and output.name == "stderr"
            ):
                return True
        return False

    def _get_failed_cells(self, nb: NotebookNode) -> List[NotebookNode]:
        failed = []
        for cell in nb.cells:
            if self._should_skip_cell(cell):
                continue

            # if it's a grade cell, the check the grade
            if utils.is_grade(cell):
                score, max_score = utils.determine_grade(cell, self.log)

                if is_extra_cell(cell):
                    score = self.validate_extra_cell(cell)

                # it's a markdown cell, so we can't do anything
                if score is not None and score < max_score:
                    failed.append(cell)
            elif (
                self.validate_all
                and cell.cell_type == "code"
                and self.code_cell_errored(cell)
            ):
                failed.append(cell)

        return failed

    def _should_skip_cell(self, cell: NotebookNode) -> bool:
        return not (self.validate_all or utils.is_grade(cell) or utils.is_locked(cell))



import base64
import pickle
import re

from nbgrader.preprocessors import NbGraderPreprocessor


class Unscramble(NbGraderPreprocessor):
    def __init__(self, **kw):
        self.__pattern = re.compile("{{([^{]+)}}")
        self.log.info("Init Unscramble")

    def preprocess(self, nb, resources):
        if "scramble_config" in nb.metadata:
            byte_config = nb.metadata.scramble_config.config
            config = pickle.loads(base64.b85decode(byte_config))

            for cell in nb.cells:
                matches = self.__pattern.findall(cell.source)
                for m in matches:
                    if m.strip() in config:
                        cell.source = cell.source.replace(
                            "{{" + m + "}}", str(config[m.strip()])
                        )

        return nb, resources



import json
import os
import traceback

from e2xcore.handlers import E2xHandler
from jupyter_core.paths import jupyter_config_path
from nbgrader.apps import NbGrader
from nbgrader.nbgraderformat import SchemaTooNewError, SchemaTooOldError
from nbgrader.server_extensions.validate_assignment.handlers import (
    NbGraderVersionHandler,
)
from tornado import web

from .validator import E2XValidator

static = os.path.join(os.path.dirname(__file__), "static")


class ValidateAssignmentHandler(E2xHandler):
    @property
    def root_dir(self):
        return self.settings["root_dir"]

    def load_config(self):
        paths = jupyter_config_path()
        paths.insert(0, os.getcwd())

        app = NbGrader()
        app.config_file_paths.append(paths)
        app.load_config_file()

        return app.config

    def validate_notebook(self, path):
        fullpath = os.path.join(self.root_dir, path)

        try:
            config = self.load_config()
            validator = E2XValidator(config=config)
            result = validator.validate(fullpath)

        except SchemaTooOldError:
            self.log.error(traceback.format_exc())
            msg = (
                "The notebook '{}' uses an old version "
                "of the nbgrader metadata format. Please **back up this "
                "notebook** and then update the metadata using:\n\nnbgrader update {}\n"
            ).format(fullpath, fullpath)
            self.log.error(msg)
            retvalue = {"success": False, "value": msg}

        except SchemaTooNewError:
            self.log.error(traceback.format_exc())
            msg = (
                "The notebook '{}' uses a newer version "
                "of the nbgrader metadata format. Please update your version of "
                "nbgrader to the latest version to be able to use this notebook."
            ).format(fullpath)
            self.log.error(msg)
            retvalue = {"success": False, "value": msg}

        except Exception:
            self.log.error(traceback.format_exc())
            retvalue = {"success": False, "value": traceback.format_exc()}

        else:
            retvalue = {"success": True, "value": result}

        return retvalue

    @web.authenticated
    def post(self):
        output = self.validate_notebook(self.get_argument("path"))
        self.finish(json.dumps(output))


# -----------------------------------------------------------------------------
# URL to handler mappings
# -----------------------------------------------------------------------------

default_handlers = [
    (r"/assignments/validate", ValidateAssignmentHandler),
    (r"/nbgrader_version", NbGraderVersionHandler),
]



import unittest

from e2xcore.utils.nbgrader_cells import grade_id
from nbformat.v4 import new_notebook

from e2xgrader.preprocessors import FilterCellsById
from e2xgrader.tests.test_utils.cells import (
    new_autograded_code_cell,
    new_autograder_test_cell,
    new_manually_graded_code_cell,
    new_readonly_markdown_cell,
)


class TestFilterCellsById(unittest.TestCase):
    def setUp(self):
        self.nb = new_notebook()
        self.nb.cells.append(
            new_readonly_markdown_cell(source="", grade_id="Task1_Task2_Description")
        )
        self.nb.cells.append(new_manually_graded_code_cell(source="", grade_id="Task1"))
        self.nb.cells.append(
            new_readonly_markdown_cell(source="", grade_id="Task2_Description")
        )
        self.nb.cells.append(new_autograded_code_cell(source="", grade_id="Task2"))
        self.nb.cells.append(new_autograder_test_cell(source="", grade_id="test_Task2"))
        self.nb.cells.append(new_readonly_markdown_cell(source="", grade_id="Task3"))
        self.grade_ids = [grade_id(cell) for cell in self.nb.cells]

    def test_no_keyword(self):
        processed_nb, _ = FilterCellsById().preprocess(self.nb, {})
        processed_cell_ids = [grade_id(cell) for cell in processed_nb.cells]
        assert all([idx in processed_cell_ids for idx in self.grade_ids])

    def test_empty_keyword(self):
        processed_nb, _ = FilterCellsById().preprocess(self.nb, {"keyword": ""})
        processed_cell_ids = [grade_id(cell) for cell in processed_nb.cells]
        assert all([idx in processed_cell_ids for idx in self.grade_ids])

    def test_non_empty_keyword(self):
        keyword = "Task2"
        processed_nb, _ = FilterCellsById().preprocess(self.nb, dict(keyword=keyword))
        processed_cell_ids = [grade_id(cell) for cell in processed_nb.cells]
        assert len(processed_cell_ids) == 4
        assert all([keyword in idx for idx in processed_cell_ids])



from tempfile import TemporaryDirectory

import nbformat
from e2xauthoring.managers import PresetManager, TemplateManager
from nbgrader.coursedir import CourseDirectory
from nbgrader.utils import is_grade


def create_temp_course():
    tmp_dir = TemporaryDirectory()
    coursedir = CourseDirectory()
    coursedir.root = tmp_dir.name
    return tmp_dir, coursedir


def add_template_with_header(coursedir, name, header_source):
    templatemodel = TemplateManager(coursedir)
    presetmodel = PresetManager(coursedir)
    res = templatemodel.new(name=name)

    nb = nbformat.read(res["path"], as_version=nbformat.NO_CONVERT)
    header_cells = presetmodel.get_template_preset("Header")
    header_cells[0].source = header_source
    nb.cells = header_cells
    nbformat.write(nb, res["path"])
    return res["path"]


def add_header_to_template(coursedir, path, header_source):
    presetmodel = PresetManager(coursedir)
    nb = nbformat.read(path, as_version=nbformat.NO_CONVERT)
    header_cells = presetmodel.get_template_preset("Header")
    header_cells[0].source = header_source
    nb.cells.extend(header_cells)
    nbformat.write(nb, path)
    return path


def add_footer_to_template(coursedir, path, footer_source):
    presetmodel = PresetManager(coursedir)
    nb = nbformat.read(path, as_version=nbformat.NO_CONVERT)
    footer_cells = presetmodel.get_template_preset("Footer")
    footer_cells[0].source = footer_source
    nb.cells.extend(footer_cells)
    nbformat.write(nb, path)
    return path


def add_question_to_task(coursedir, path, question_type, grade_id=None, points=0):
    presetmodel = PresetManager(coursedir)
    nb = nbformat.read(path, as_version=nbformat.NO_CONVERT)
    cells = presetmodel.get_question_preset(question_type)
    if grade_id is not None:
        for cell in cells:
            cell.metadata.nbgrader.grade_id = cell.metadata.nbgrader.grade_id.replace(
                "task", grade_id
            )
            if is_grade(cell):
                cell.metadata.nbgrader.points = points
    nb.cells.extend(cells)
    nbformat.write(nb, path)
    return path
